# ***************************************************************************
# Copyright 2020 VMware, Inc.  All rights reserved. VMware Confidential.
# ***************************************************************************
# $Id$
# $DateTime$
# $Change$
# $Author$
# ***************************************************************************
import json
import nsx_logging
import os

from cli.commands.cli_command_provider import CliCommandProvider
from cli.commands.common.arguments import ApiThumbprintArgument
from cli.utils.apiclient import request as nsx_api_request
from cli.utils.constants import NODE_TYPE_EDGE
from cli.utils.constants import NODE_TYPE_PUBLIC_CLOUD_GATEWAY
from cli.utils.error_utils import get_message
from cli.utils.exceptions import ApiException
from cli.utils.ipaddr_utils import split_address
from cli.utils.location_utils import FilePaths
from cli.utils.mpa_utils import get_mpa_node_config
from cli.utils.node_utils import get_node_type

lg = nsx_logging.getLogger("cli.commands.common.push_certificate")

MGR_SUCCESS = 0
PUSH_CERTIFICATE_FAILED = 1

_MP_API_PORT = 443
_HV_CLIENT_TYPE = "cvn-hv"
_EDGE_CLIENT_TYPE = "cvn-edge"

def _pushCertificate(api_server, api_username, api_thumbprint, api_password):
    if os.path.isfile(FilePaths.HV_CERT_FILE):
        host_cert = open(FilePaths.HV_CERT_FILE, 'r').read()
    else:
        err_msg = "Host certificate not found"
        return PUSH_CERTIFICATE_FAILED, err_msg

    if not os.access(FilePaths.MPA_CONFIG_FILE, os.W_OK):
        err_msg = "MPA config file does not exist"
        return PUSH_CERTIFICATE_FAILED, err_msg
    try:
        mpaconfig_node_uuid, account_name, shared_secret = get_mpa_node_config()
    except ValueError:
            err_msg = "MPA config file is corrupted"
            return PUSH_CERTIFICATE_FAILED, err_msg

    reg_method = "PUT"
    reg_uri = "/nsxapi/api/v1/messaging/clients/%s" % mpaconfig_node_uuid
    api_ip_or_host, api_port = split_address(api_server)
    api_port = _MP_API_PORT if api_port is None else api_port
    if get_node_type() in [NODE_TYPE_EDGE, NODE_TYPE_PUBLIC_CLOUD_GATEWAY]:
        client_type = _EDGE_CLIENT_TYPE
    else:
        client_type = _HV_CLIENT_TYPE

    reg_body = json.dumps({
       "client_type" : client_type,
       "client_token" : mpaconfig_node_uuid,
       "client_id" : mpaconfig_node_uuid,
       "account_name" : account_name,
       "certificate": host_cert,
       "account_shared_secret": shared_secret
    })
    headers = {"Content-Type": "application/json"}
    try:
        status, json_resp, text, _ = nsx_api_request(api_ip_or_host, api_port,
                                                     api_thumbprint,
                                                     {"username": api_username, "password": api_password},
                                                     reg_method,
                                                     reg_uri, body=reg_body,
                                                     headers=headers)
    except ApiException as ex:
        return PUSH_CERTIFICATE_FAILED, get_message(ex.args[0])

    return MGR_SUCCESS, None


class PushHostCertificate(CliCommandProvider):
    def __call__(self, cmd_list):
        username = self.get_arg_value("api-username", cmd_list)
        thumbprint = ApiThumbprintArgument.normalize(
            self.get_arg_value("api-thumbprint", cmd_list))
        password = self.get_arg_value("api-password", cmd_list)

        # If password isn't entered by the user with the rest of the push
        # command, the framework prompts for it in which case the password
        # value is guaranteed to be the last argument value.
        if len(cmd_list) > 8:
            password = self.get_arg_value("api-password", cmd_list)
        else:
            password = cmd_list[-1]

        result, err_msg = _pushCertificate(cmd_list[2], username,
                                               thumbprint, password)
        if result == PUSH_CERTIFICATE_FAILED:
            err_msg = err_msg if err_msg else "internal error"
            return "PUSH_CERTIFICATE_FAILED\n%s" % err_msg

        return "OK\n\n%s" % json.dumps("Host certificate was pushed to management plane successfully")
