# **********************************************************************
# Copyright 2016-2017 VMware, Inc.  All rights reserved. VMware Confidential.
# **********************************************************************
# $Id$
# $DateTime$
# $Change$
# $Author$
# **********************************************************************
import nsx_logging
try:
    import simplejson
except ImportError:
    import json as simplejson
import time
import fcntl
import struct

from cli.commands.cli_command_provider import CliCommandProvider
from cli.commands.common.arguments import get_interface_names
import cli.commands.host_shared.ovs_util as ovs_util
from cli.utils.command_utils import invoke_command
from cli.utils.error_code import ErrorCode as ec
from cli.utils.pcap_utils import get_nsx_agent_style_direction

lg = nsx_logging.getLogger("cli.commands.host_kvm.ops.packet_capture")

IFF_TAP = 0x0002
IFF_NO_PI = 0x1000
TUNSETIFF = 0x400454ca
TUNSETOWNER = TUNSETIFF + 2


class PacketCaptureCommandRunner(object):
    default_span_session = "0"
    default_span_port = "span-%s" % default_span_session
    default_monitor_direction = "both"
    sessions = {}

    def run(self, cmd_list):
        rc, out, err = invoke_command(cmd_list, display=True)
        if rc != 0:
            lg.error("Failed to span session: cmd '%s', err: '%s'",
                     " ".join(cmd_list), err,
                     error_code=ec.OPERATION_FAILED)
            return "MSG_CLI_UNEXPECTED_ERROR\n%s" % err
        if not out:
            out = simplejson.dumps("")
        return "OK\n\n%s" % out

    # Check to see if the span_port is up as local interface
    def check_span_port(self, span_port):
        return span_port in get_interface_names(include_down=True)

    # Creating the default span session in KVM for packet capture
    def create_default_kvm_span_session(self, lport, direction):
        rv = self.create_kvm_span_session(self.default_span_port)
        nsx_agent_style_direction = get_nsx_agent_style_direction(direction)
        if rv.startswith("OK"):
            # mapped vif into lport if it is
            is_vif, mapped_lport = ovs_util.get_lport_on_vif(lport)
            if is_vif:
               lport = mapped_lport
            cmds = [self.default_span_session, lport, nsx_agent_style_direction]
            # Installing flow in ovs for forwarding packets to the span_port
            rv = ovs_util.set_capture_lport(cmds)
            if not rv:
                out = simplejson.dumps("")
                rv = "OK\n\n%s" % out
            else:
                # For the default session (span 0), delete the session in
                # case of any error while installing flow in ovs
                self.destroy_default_kvm_span_session()
            return rv
        else:
            self.destroy_kvm_span_session(self.default_span_port)
            return rv

    # Creating a span session in KVM for packet capture
    def create_kvm_span_session(self, span_port):
        return self.create_span_port(span_port)

    # Creating the span_port in KVM
    def create_span_port(self, span_port):
        if not self.check_span_port(span_port):
            rv = self.run(["ip", "tuntap", "add", "dev", span_port,
                           "mode", "tap"])
            time.sleep(1)
            if rv.startswith("OK"):
                return self.set_port_status(span_port)
            rv = "MSG_CLI_UNEXPECTED_ERROR\nFailed to create span port"
            return rv
        out = simplejson.dumps("")
        rv = "OK\n\n%s" % out
        return rv

    # Setting the span_port to up
    def set_port_status(self, span_port):
        rv = self.run(["ip", "link", "set", span_port, "up"])
        time.sleep(1)
        if rv.startswith("OK"):
            return self.add_span_port_to_ovs(span_port)
        rv = "MSG_CLI_UNEXPECTED_ERROR\nFailed to set the status of port"
        return rv

    # Adding the span port to ovs
    def add_span_port_to_ovs(self, span_port):
        # Find the bridge name with external id nsx-managed
        cmd = ["ovs-vsctl", "--columns=name", "find", "bridge",
                "external_ids={bridge-id=nsx-managed}"]
        rc, out, err = invoke_command(cmd, False)
        if (rc != 0 or (not out)):
            lg.error("Couldn't find the bridge-name: cmd '%s', err: '%s'",
                     " ".join(cmd), err, error_code=ec.OPERATION_FAILED)
            return "MSG_CLI_UNEXPECTED_ERROR\nFailed to create span port"
        # Extracting the bridge name from the output string
        nsx_bridge = out.split(": ")[1]
        rv = self.run(["ovs-vsctl", "add-port", nsx_bridge.split("\n")[0],
                      span_port])
        time.sleep(1)
        if rv.startswith("OK"):
            rv = "OK\n\n%s" % simplejson.dumps("")
        else:
            rv = "MSG_CLI_UNEXPECTED_ERROR\nFailed to add span port to ovs"
        return rv

    # Open the connection to span_port
    def open_connection_to_port(self, span_port):
        rc = False
        try:
            tun = open('/dev/net/tun', 'r+b')
            self.sessions[span_port] = tun
            ifr = struct.pack('16sH', span_port.encode('utf-8'),
                              IFF_TAP | IFF_NO_PI)
            fcntl.ioctl(tun, TUNSETIFF, ifr)
            fcntl.ioctl(tun, TUNSETOWNER, 1000)
            # Successful connection to the span_port will assign an IP address
            # to it. Deleting the IP address to avoid neighbor discovery
            # packet exchange for the span_port.
            # In case of an exception while opening the connection, False will
            # be returned, and the calling method will delete the span_port.
            self.delete_ip_address_from_span_port(span_port)
            rc = True
            return rc
        except:
            return rc

    # After opening connection to span_port, an IP address gets assigned to it
    # Deleting the IP addr to avoid neighbor discovery packets for span_port
    def delete_ip_address_from_span_port(self, span_port):
        self.run(["ip", "addr", "flush", "dev", span_port])

    # Destroying the default (span 0) span session
    def destroy_default_kvm_span_session(self):
        # For the default (span 0) session, delete the flow from ovs
        self.unmonitor_default_kvm_session()

        # Destroy the port and capture session
        self.destroy_kvm_span_session(self.default_span_port)

    # Destroying the KVM span session
    def destroy_kvm_span_session(self, span_port):
        #Removing the span_port from ovs
        self.delete_span_port_from_ovs(span_port)

        #Deleting the span_port
        self.delete_span_port(span_port)

    # Delete span_port from ovs if it exist
    def delete_span_port_from_ovs(self, span_port):
        if self.check_span_port(span_port):
            # Find the bridge name with external id nsx-managed
            cmd = ["ovs-vsctl", "--columns=name", "find", "bridge",
                   "external_ids={bridge-id=nsx-managed}"]
            rc, out, err = invoke_command(cmd, False)
            if (rc != 0 or (not out)):
                lg.error("Couldn't find the bridge-name: cmd '%s', err: '%s'",
                         " ".join(cmd), err, error_code=ec.OPERATION_FAILED)
                return "MSG_CLI_UNEXPECTED_ERROR\nFailed to delete span port"
            # Extracting the bridge name from the output string
            nsx_bridge = out.split(": ")[1]
            self.run(["ovs-vsctl", "del-port", nsx_bridge.split("\n")[0],
                     span_port])

    # Delete span_port
    def delete_span_port(self, span_port):
        # set delete the tuntap forcely or it raises an issue described in
        # PR2052531
        self.run(["ip", "link", "delete", span_port])

    # Close the connection to span_port
    def close_connection_to_port(self, span_port):
        if span_port in self.sessions:
            self.sessions[span_port].close()

    # Unmonitor the default KVM session
    def unmonitor_default_kvm_session(self):
        cmd = [self.default_span_session]
        ovs_util.del_capture_session(cmd)


class MonitorPacketCaptureSessionPort(CliCommandProvider):
    def __call__(self, cmd_list):
        kvm_session = PacketCaptureCommandRunner()
        session_id = cmd_list[3]
        span_port = "span-%s" % session_id
        lport = cmd_list[5]
        direction = cmd_list[7]
        nsx_agent_style_direction = get_nsx_agent_style_direction(direction)
        rv = kvm_session.create_kvm_span_session(span_port)
        if rv.startswith("OK"):
            # mapped vif into lport if it is
            is_vif, mapped_lport = ovs_util.get_lport_on_vif(lport)
            if is_vif:
               lport = mapped_lport
            cmds = [session_id, lport, nsx_agent_style_direction]
            # Installing flow in ovs for forwarding packets to the span_port
            rv = ovs_util.set_capture_lport(cmds)
            if not rv:
                out = simplejson.dumps("")
                rv = "OK\n\n%s" % out
        else:
            # Destroying the session in case of any error while creating
            kvm_session.destroy_kvm_span_session(span_port)
        return rv


class ShowPacketCaptureSession(CliCommandProvider):
    def __call__(self, cmd_list):
        cmds = []
        ret_value = []
        sessions = {}
        if len(cmd_list) > 3:
            cmds.append(cmd_list[3])
            sessions[int(cmd_list[3])] = []
        else:
            for session_id in range(0, 4):
                sessions[session_id] = []
        out = ovs_util.get_capture_sessions(cmds)
        # Populating the dictionary with session_id(key) and lports(value)
        for lst in out:
            sessions[int(lst[0])].append(lst[1])
        # Creating a list of sessions
        for s_id in sessions:
            # skip session 0, which is used by "start capture interface" only
            if s_id == 0:
               continue
            ret = {}
            ret['ID'] = s_id
            ret['PORTS'] = sessions[s_id]
            ret['VIFS'] = [ovs_util.get_lport_vif(lport) for lport in sessions[s_id]]
            ret_value.append(ret)
        return "OK\n\n%s" % simplejson.dumps(ret_value)


class UnMonitorPacketCaptureSessionPort(CliCommandProvider):
    def __call__(self, cmd_list):
        session_id = cmd_list[3]
        lport = cmd_list[5]
        # mapped vif into lport if it is
        is_vif, mapped_lport = ovs_util.get_lport_on_vif(lport)
        if is_vif:
           lport = mapped_lport
        cmds = [session_id, lport]
        rv = ovs_util.del_capture_lport(cmds)
        if not rv:
            out = simplejson.dumps("")
            return "OK\n\n%s" % out
        else:
            return "MSG_CLI_UNEXPECTED_ERROR\nFailed to delete the lport"


class UnMonitorPacketCaptureSession(CliCommandProvider,
                                    PacketCaptureCommandRunner):
    def __call__(self, cmd_list):
        kvm_session = PacketCaptureCommandRunner()
        span_port = "span-%s" % cmd_list[3]
        session_id = cmd_list[3]
        cmds = [session_id]
        rv = ovs_util.del_capture_session(cmds)
        if not rv:
            out = simplejson.dumps("")
            # Deleting the span session in case of successful deletion from ovs
            kvm_session.destroy_kvm_span_session(span_port)
            return "OK\n\n%s" % out
        else:
            return "MSG_CLI_UNEXPECTED_ERROR\nFailed to delete the session"
