# ***************************************************************************
# Copyright 2014-2020 VMware, Inc.  All rights reserved. VMware Confidential.
# ***************************************************************************
# $Id$
# $DateTime$
# $Change$
# $Author$
# ***************************************************************************
from hashlib import sha256
import json
import nsx_logging
import os
import socket
import ssl
import string
import subprocess
import time
import uuid
import re
import tempfile
from xml.dom import minidom
import xml.etree.ElementTree as ET

from cli.commands.host_shared.aph_info_config import (
    _write_appliance_info_xml as write_appliance_info_xml,
    _get_aph_info_from_listener as get_aph_info,
)
from cli.commands.cli_command_provider import CliCommandProvider
from cli.commands.common.arguments import HostnameOrIpAddressArgument
from cli.commands.common.arguments import ApiThumbprintArgument
from cli.utils.command_utils import invoke_command
from cli.utils.apiclient import request as nsx_api_request
from cli.utils.constants import NODE_TYPE_HOST_CENTOSKVM
from cli.utils.constants import NODE_TYPE_HOST_SLESKVM
from cli.utils.constants import NODE_TYPE_HOST_ESX
from cli.utils.constants import NODE_TYPE_HOST_RHELKVM
from cli.utils.constants import NODE_TYPE_HOST_UBUNTUKVM
from cli.utils.constants import NODE_TYPE_HOST_WIN
from cli.utils.constants import NODE_TYPE_BAREMETAL_UBUNTU_SERVER
from cli.utils.constants import NODE_TYPE_BAREMETAL_RHEL_SERVER
from cli.utils.constants import NODE_TYPE_BAREMETAL_CENTOS_SERVER
from cli.utils.constants import NODE_TYPE_BAREMETAL_SLES_SERVER
from cli.utils.constants import NODE_TYPE_BAREMETAL_OEL_SERVER
from cli.utils.constants import NODE_TYPE_BAREMETAL_WINDOWS_SERVER
from cli.utils.constants import NODE_TYPE_BAREMETAL_RHEL_CONTAINER
from cli.utils.constants import NODE_TYPE_BAREMETAL_CENTOS_CONTAINER
from cli.utils.error_utils import get_message
from cli.utils.exceptions import ApiException
from cli.utils.ipaddr_utils import split_address
from cli.utils.cert_utils import create_self_signed_cert
from cli.utils.cert_utils import get_cert_subject
from cli.utils.location_utils import FilePaths
from cli.utils.mpa_utils import get_mpa_node_config
from cli.utils.node_utils import get_node_type, is_windows, is_reboot_required
from cli.utils.error_code import ErrorCode as ec
from compat23 import compat23_httplib

lg = nsx_logging.getLogger("cli.commands.host_shared.register")


node_type = get_node_type()
if node_type == NODE_TYPE_HOST_ESX:
    from esx_utils import get_management_ip_addresses
    from esx_utils import get_os_version
elif node_type in (NODE_TYPE_BAREMETAL_CENTOS_CONTAINER,
                   NODE_TYPE_BAREMETAL_CENTOS_SERVER,
                   NODE_TYPE_BAREMETAL_RHEL_CONTAINER,
                   NODE_TYPE_BAREMETAL_RHEL_SERVER,
                   NODE_TYPE_BAREMETAL_UBUNTU_SERVER,
                   NODE_TYPE_BAREMETAL_SLES_SERVER,
                   NODE_TYPE_BAREMETAL_OEL_SERVER,
                   NODE_TYPE_HOST_CENTOSKVM,
                   NODE_TYPE_HOST_SLESKVM,
                   NODE_TYPE_HOST_RHELKVM,
                   NODE_TYPE_HOST_UBUNTUKVM):
    from kvm_utils import get_management_ip_addresses
    from kvm_utils import get_os_version
    import errno
    from pwd import getpwnam
    from grp import getgrnam
elif node_type in (NODE_TYPE_HOST_WIN,
                   NODE_TYPE_BAREMETAL_WINDOWS_SERVER):
    from win_utils import get_management_ip_addresses
    from win_utils import get_os_version


MGR_SUCCESS = 0
MGR_AUTHENTICATION_FAILED = 1
MGR_REGISTER_EXCEPTION = 2
MGR_DEREGISTER_EXCEPTION = 3
MGR_REGISTER_FAILED = 4
MGR_DEREGISTER_FAILED = 5

_MP_ACCOUNT_SECRET_LEN = 42
_MP_ACCOUNT_SECRET_ALLOWED_CHARS = string.digits + string.ascii_letters
_MP_ACCOUNT_SECRET_ALLOWED_CHARS_LEN = len(_MP_ACCOUNT_SECRET_ALLOWED_CHARS)

_MPA_ACCOUNT_NAME_KEY = "AccountName"
_MPA_SHARED_SECRET_KEY = "SharedSecret"
_MPA_CLIENT_TYPE_KEY = "RmqClientType"
_MPA_BROKER_CLUSTER_KEY = "RmqBrokerCluster"
_MPA_BROKER_IS_MASTER_KEY = "BrokerIsMaster"
_MPA_BROKER_VIRTUAL_HOST_KEY = "BrokerVirtualHost"
_MPA_BROKER_IP_ADDR_KEY = "BrokerIpAddress"
_MPA_BROKER_FQDN_KEY = "BrokerFqdn"
_MPA_BROKER_PORT_KEY = "BrokerPort"
_MPA_BROKER_SSL_TP_KEY = "BrokerSslCertThumbprint"

_MP_API_PORT = 443
_HV_CLIENT_TYPE = "cvn-hv"

_MSG_HOST_CANCEL_DELETE_NSX="NSX environment deletion on host is cancelled."
_MSG_HOST_SUCCESS_DELETE_NSX="NSX environment on host deleted successfully"


def _generate_account_secret():
    try:
        from OpenSSL import rand  # Import here so api-doc-gen won't import it
        secret_bytes = rand.bytes(_MP_ACCOUNT_SECRET_LEN)
    except (ImportError, AttributeError):
        if is_windows():
            openssl_cmd = FilePaths.OPENSSL_PATH
        else:
            openssl_cmd = "openssl"
        secret_bytes = subprocess.check_output(
            [openssl_cmd, "rand", "%s" % _MP_ACCOUNT_SECRET_LEN])
    secret = []
    for byte in secret_bytes:
        byte = ord(byte) if isinstance(byte, str) else byte
        secret.append(_MP_ACCOUNT_SECRET_ALLOWED_CHARS[
            byte % _MP_ACCOUNT_SECRET_ALLOWED_CHARS_LEN])
    return "".join(secret)


def _nsxproxy_service(is_start):
    """Restart nsxproxy service
    """

    if is_start:
        start_or_stop = "start"
        start_log_str = "Starting"
    else:
        start_or_stop = "stop"
        start_log_str = "Stopping"
    lg.info(start_log_str + " nsx-proxy")

    if is_windows():
        command_nsxproxy = ["sc", start_or_stop, "nsx-proxy"]
    else:
        if os.path.isfile(FilePaths.NSX_PROXY_INIT_SCRIPT):
           command_nsxproxy = [FilePaths.NSX_PROXY_INIT_SCRIPT, start_or_stop]

    if command_nsxproxy is not None:
        rc, p1_out, p1_err = invoke_command(command_nsxproxy)
        if rc != 0:
            lg.warning("Error in " + start_log_str.lower() +
                       " nsx-proxy stdout: %s stderr: %s",
                       p1_out, p1_err);
            return False
    return True


def _mpa_services(is_start):
    if is_start:
        start_or_stop = "start"
        start_log_str = "Starting"
    else:
        start_or_stop = "stop"
        start_log_str = "Stopping"
    lg.info(start_log_str + " nsx-mpa to update mpaconfig")
    command_mpa = None
    if is_windows():
        command_mpa = None
    else:
        if os.path.isfile(FilePaths.MPA_INIT_SCRIPT):
            command_mpa = [FilePaths.MPA_INIT_SCRIPT, start_or_stop]

    if command_mpa is not None:
        rc, p2_out, p2_err = invoke_command(command_mpa)
        if rc != 0:
            lg.warning("Error in " + start_log_str.lower() +
                       " nsx-mpa stdout: %s stderr: %s",
                       p2_out, p2_err);


def _write_mpa_config(account_name, shared_secret, brokers,mp_endpoint='', thumbrint='', token='', nodeUUId='', restart_services=True):
    if not os.path.isfile(FilePaths.MPA_CONFIG_FILE):
        # MPA config file should exist with sticky bit
        # as part of nsx-mpa VIB install
        lg.warning("MPA config file does not exist")
        return -1
    mpa_config_body = {
        _MPA_ACCOUNT_NAME_KEY: account_name,
        _MPA_SHARED_SECRET_KEY: shared_secret,
        _MPA_CLIENT_TYPE_KEY: _HV_CLIENT_TYPE,
        _MPA_BROKER_CLUSTER_KEY: brokers
    }
    try:
        if restart_services:
            _mpa_services(False)
        with open(FilePaths.MPA_CONFIG_FILE, "w") as fp:
            json.dump(mpa_config_body, fp, indent=4)

    finally:
        if restart_services:
            _mpa_services(True)
    return 0


def _write_appliance_info_xml(aph_list=[], restart_services=True):
    ret_val = True
    try:
        if restart_services:
            _mpa_services(False)
            _nsxproxy_service(False)
        ret_val = write_appliance_info_xml(aph_list)
    finally:
        if restart_services:
            _nsxproxy_service(True)
            _mpa_services(True)
    return ret_val


def _aphconfig_update_services_action(is_start):
    services = [_nsxproxy_service, _mpa_services]
    if not is_start:
        services.reverse()
    for service in services:
        service(is_start)


def _should_update_host_cert(node_uuid):
    """Check if the uuid in host-cfg.xml is same as node_uuid
    """

    sn = get_cert_subject(FilePaths.HV_CERT_FILE)
    r1 = re.match('.*UID\s*=\s*(.*)$', sn)
    if r1 and node_uuid == r1.group(1):
        return False
    return True


def _gen_host_cert(node_uuid):
    """Read the contents of host-cfg.xml
    """

    PRIVKEY_FILE = FilePaths.HV_PRIVKEY_FILE
    CERT_FILE = FilePaths.HV_CERT_FILE
    KEY_CERT_DIR = FilePaths.NSX_COMMON_DIR

    key_file = None
    cert_file = None

    # Check if the key and cert exist and if we have write access to
    # both of them. We won't modify these files here, but while we
    # persist mpaconfig / appliance info.
    if os.path.exists(PRIVKEY_FILE) and not os.access(PRIVKEY_FILE, os.W_OK):
            lg.warning("No write permissions for host key")
            return False, None, None
    if os.path.exists(CERT_FILE) and not os.access(CERT_FILE, os.W_OK):
            lg.warning("No write permissions for host cert")
            return False, None, None

    # If neither key or cert exist, check if we have write access to the dir
    if not os.path.exists(PRIVKEY_FILE) and not os.path.exists(CERT_FILE):
        if not os.access(KEY_CERT_DIR, os.W_OK):
            lg.warning("No write permissions for host key/cert dir")
            return False, None, None

    # Create temp key file, cert file and openssl conf file
    try:
        key_file = tempfile.mktemp()
        cert_file = tempfile.mktemp()
        tmp_openssl_cnf_file = tempfile.mktemp()
    except Exception:
        lg.exception("Error creating temp files for cert creation",
            error_code=ec.INTERNAL_ERROR)
        return False, None, None

    try:
        tmp_openssl_cnf_filep = open(tmp_openssl_cnf_file, "w")
        openssl_cnf_lines = [
            "[req]\n",
            "default_bits            = 2048\n",
            "distinguished_name      = req_distinguished_name\n",
            "#Don't encrypt the key\n",
            "encrypt_key             = no\n",
            "prompt                  = no\n",
            "string_mask             = nombstr\n",
            "\n",
            "[ req_distinguished_name ]\n",
            "countryName             = US\n",
            "stateOrProvinceName     = California\n",
            "localityName            = Palo Alto\n",
            "0.organizationName      = VMware, Inc.\n",
            "emailAddress            = ssl-certificates@vmware.com\n",
            "commonName              = VMware-NSX-Host\n",
            "UID                     = " + node_uuid + "\n",
            "\n",
            "[ req_ext ]\n",
            "basicConstraints        = CA:FALSE\n",
            "extendedKeyUsage        = clientAuth\n",
            "subjectAltName          = @alt_names\n",
            "\n",
            "[ alt_names ]\n",
            "URI                     = nsx-tn:" + node_uuid + "\n"]

        tmp_openssl_cnf_filep.writelines(openssl_cnf_lines)
        tmp_openssl_cnf_filep.close()
        create_self_signed_cert(tmp_openssl_cnf_file, key_file, cert_file,
                                "rsa:2048", 3650)
    finally:
        try:
            os.remove(tmp_openssl_cnf_file)
        except Exception:
            # Just log the error. Not a failure condition.
            lg.exception("Error cleaning up temp openssl config",
                error_code=ec.INTERNAL_ERROR)

    return True, key_file, cert_file


def _read_host_cfg_xml():
    """Read the contents of host-cfg.xml
    """
    try:
        tree = ET.parse(FilePaths.HOST_CFG_FILE)
    except Exception as ex:
        lg.exception("Error reading host-cfg.xml",
                     error_code=ec.INTERNAL_ERROR)
        return None
    return tree


def _write_host_cfg_xml(node_uuid, restart_service=True):
    """Persist UUID and key/cert info in host-cfg.xml

    Revisit host-cfg.xml persistence lifecycle when addressing
    changes for appliance-info.xml persistence.
    """

    if not os.path.isfile(FilePaths.HOST_CFG_FILE):
        # host-cfg.xml file should exist as part of nsx-proxy install
        lg.warning("host-cfg.xml does not exist")
        return False

    config = ET.Element('config')
    tn_uuid  = ET.SubElement(config, 'uuid')
    host_private_key_file = ET.SubElement(config, 'hostPrivateKeyFile')
    host_cert_file = ET.SubElement(config, 'hostCertificateFile')

    host_private_key_file.text = FilePaths.HV_PRIVKEY_FILE
    host_cert_file.text = FilePaths.HV_CERT_FILE
    tn_uuid.text = node_uuid

    mydata = ET.tostring(config)
    xmlstr = minidom.parseString(
        ET.tostring(config)).toprettyxml(indent=' ' * 4)

    ret_val = True
    try:
        if restart_service:
            ret = _nsxproxy_service(False)
            if not ret:
                return False

        with open(FilePaths.HOST_CFG_FILE, "w") as fp:
            fp.write(xmlstr)
    except Exception:
        lg.exception("Error updating host-cfg.xml",
                     error_code=ec.INTERNAL_ERROR)
        ret_val = False
    finally:
        if restart_service:
            _nsxproxy_service(True)
    return ret_val


def _get_host_cert_key():
    success = True
    host_cert = None
    host_key = None

    try:
        host_cert = open(FilePaths.HV_CERT_FILE, 'r').read()
        host_key = open(FilePaths.HV_PRIVKEY_FILE, 'r').read()
    except Exception:
        lg.exception("Error reading host cert/key",
            error_code=ec.INTERNAL_ERROR)
        success = False

    return success, host_cert, host_key


def _update_host_cert_key(cert, key):
    cert_file = None
    key_file = None
    PRIVKEY_FILE = FilePaths.HV_PRIVKEY_FILE
    CERT_FILE = FilePaths.HV_CERT_FILE
    NSX_PROXY_USER = "nsx-proxy"
    NSX_PROXY_GROUP = "nsx-proxy"
    uid = gid = 0

    change_file_metadata = True
    if os.path.exists(CERT_FILE) and os.path.exists(PRIVKEY_FILE):
        change_file_metadata = False
    try:
        with open(CERT_FILE, "w") as cert_file:
            cert_file.write(cert)
        with open(PRIVKEY_FILE, "w") as key_file:
            key_file.write(key)
    except Exception:
        lg.exception("Error updating host cert/key",
            error_code=ec.INTERNAL_ERROR)
        return False

    if change_file_metadata:
        try:
            if node_type != NODE_TYPE_HOST_ESX:
                os.chmod(CERT_FILE, 0o660)
                os.chmod(PRIVKEY_FILE, 0o660)
                try:
                    uid = getpwnam(NSX_PROXY_USER).pw_uid
                    gid = getgrnam(NSX_PROXY_GROUP).gr_gid
                    os.chown(CERT_FILE, uid, gid)
                    os.chown(PRIVKEY_FILE, uid, gid)
                except KeyError:
                    lg.warning("Could changing cert user/group ownership")
            else:
                os.chmod(CERT_FILE, 0o644)
                os.chmod(PRIVKEY_FILE, 0o644)
        except Exception:
            lg.exception("Error updating ownership/permissions of host cert/key",
                error_code=ec.INTERNAL_ERROR)
    return True


def _backup_host_state(old_host_state):
    # Backup node uuid in host-cfg.xml. If host-cfg.xml is empty/invalid,
    # it is ok to update it with an empty UUID during restore.
    old_uuid = ''
    host_cfg_tree = _read_host_cfg_xml()
    if host_cfg_tree:
        old_uuid = host_cfg_tree.findtext('uuid', '')
        old_host_state['uuid'] = old_uuid

    # Backup node cert/key
    success, host_cert, host_key = _get_host_cert_key()
    if success:
        old_host_state['host_cert'] = host_cert
        old_host_state['host_key'] = host_key


def _restore_host_state(host_state):
    if 'uuid' in host_state:
        _write_host_cfg_xml(host_state['uuid'])

    if 'host_cert' in host_state and 'host_key' in host_state:
        _update_host_cert_key(host_state['host_cert'], host_state['host_key'])


def _register(api_server, api_username, api_thumbprint, api_password, token='', tokenfile='', node_id=''):
    """Register node by invoking configured manager(s)' northbound API.
    There are two workflows involved based on the arguments specified.
    New Node registration workflow when NodeId is not specified :
      - Generate a new Node ID
      - Create payload from scratch with generated NodeId
      - POST /api/v1/fabric/nodes?action=register_node
    Node registration workflow for an existing Node when NodeId and
    Token are specified :
      - POST /api/v1/fabric/nodes/<node-id>?action=register_node
    Args:
        api_server: Hostname or IP address and optional :port of a MP API server.
        api_thumbprint: Expected thumbprint of manager API server.
        api_username: Username to authenticate with MP API.
        api_password: Password to authenticate with MP API.
        api_token: Limited time auth token of NSX Manager.
        node_id: NodeId with which this host will be registered.
    Return:
        (MGR_SUCCESS, node_uuid, None) if registration succeeded.
        (MGR_AUTHENTICATION_FAILED, None, err_msg) tuple if authentication failed.
        (MGR_REGISTER_EXCEPTION, None, err_msg) tuple if exception is raised.
        (MGR_REGISTER_FAILED, None, err_msg) tuple in case of any other failure.
    """

    node_type = get_node_type()
    if node_type == NODE_TYPE_HOST_ESX:
        version_command_op = subprocess.check_output(['vmware','-v'])
        version_command_op = str(version_command_op.strip()).replace('\'','')
        version = version_command_op.split(' ')[2]
        buildnum = version_command_op.split(' ')[3].split('-')[1]
        lg.info("version %s buildnum %s", version, buildnum)
        if (version == '6.5.0' and int(buildnum) < 10884925) or \
           (version == '6.7.0' and int(buildnum) < 11405166):
            err_msg = "Invalid host type or host version not supported"
            return MGR_REGISTER_FAILED, None, err_msg

    node_uuid = ""
    host_cert = ""
    host_key = ""
    account_name = ""
    if not os.access(FilePaths.MPA_CONFIG_FILE, os.W_OK):
        err_msg = "MPA config file does not exist or is not writable"
        return MGR_REGISTER_FAILED, None, err_msg
    try:
        mpaconfig_node_uuid, account_name, shared_secret = get_mpa_node_config()
    except ValueError:
        mpaconfig_node_uuid = None
        shared_secret = None
        lg.exception("MPA config file is corrupted.",
                     error_code=ec.INVALID_CONFIG)
    reg_uri = "/api/v1/fabric/nodes?action=register_node"
    headers = {"Content-Type": "application/json"}

    if tokenfile:
        lg.info('tokenfile provided as ' + tokenfile)
        with open(tokenfile, 'r') as file:
            for line in file:
                token = line.replace("\r", "").replace("\n", "")

        if token:
            lg.info('Token is read from file')
        else:
            lg.info('Token is not present in the given file')

    else:
        lg.info('Tokenfile is not given')

    if node_id:
        node_uuid = node_id
        reg_uri = "/api/v1/fabric/nodes/" + node_uuid + "?action=register_node"
        headers["Authorization"] = "RegistrationToken " + token
    if not node_uuid:
        if mpaconfig_node_uuid:
            node_uuid = mpaconfig_node_uuid
        else:
            node_uuid = str(uuid.uuid1())
    account_name = _HV_CLIENT_TYPE + "-" + node_uuid

    # Backup host state to restore in case of failure
    old_host_state = {}
    _backup_host_state(old_host_state)

    get_existing_cert = True
    needs_proxy_restart = True
    if _should_update_host_cert(node_uuid):
        # Create host cert/key pair. If this fails, then we revert to the old
        # behavior of nsx-proxy creating its own key/cert.
        ret, tmp_key_file, tmp_cert_file = _gen_host_cert(node_uuid)
        if ret:
            try:
                host_cert = open(tmp_cert_file, 'r').read()
                host_key  = open(tmp_key_file, 'r').read()
                get_existing_cert = False
            except Exception:
                lg.exception("Error reading temp cert/key",
                             error_code=ec.INTERNAL_ERROR)
            # We've read the cert/key file contents already. Clean them up now.
            for path in [tmp_key_file, tmp_cert_file]:
                try:
                    os.remove(path)
                except Exception:
                    lg.exception("Error cleaning up temp key/cert file: " + path,
                         error_code=ec.INTERNAL_ERROR)
            if not get_existing_cert:
                # Write host-cfg.xml, but do not restart nsx_proxy
                result = _write_host_cfg_xml(node_uuid, False)
                if not result:
                    err_msg = "Unable to update host-cfg.xml"
                    return MGR_REGISTER_FAILED, None, err_msg
    else:
        needs_proxy_restart = False

    if get_existing_cert:
        if needs_proxy_restart:
            # The resulting nsxproxy service restart updates certificates if
            # necessary.
            result = _write_host_cfg_xml(node_uuid)
            if not result:
                err_msg = "Unable to update host-cfg.xml"
                return MGR_REGISTER_FAILED, None, err_msg
        success, host_cert, host_key = _get_host_cert_key()
        if not success:
            lg.warning("_register: Host cert not found at %s",
                       FilePaths.HV_CERT_FILE)

    if not shared_secret:
        shared_secret = _generate_account_secret()
    reg_method = "POST"
    if node_type == NODE_TYPE_HOST_CENTOSKVM:
        os_type = "CENTOSKVM"
    elif node_type == NODE_TYPE_HOST_SLESKVM:
        os_type = "SLESKVM"
    elif node_type == NODE_TYPE_HOST_ESX:
        os_type = "ESXI"
    elif node_type == NODE_TYPE_HOST_RHELKVM:
        os_type = "RHELKVM"
    elif node_type == NODE_TYPE_HOST_UBUNTUKVM:
        os_type = "UBUNTUKVM"
    elif node_type == NODE_TYPE_HOST_WIN:
        os_type = "HYPERV"
    elif node_type == NODE_TYPE_BAREMETAL_UBUNTU_SERVER:
        os_type = "UBUNTUSERVER"
    elif node_type == NODE_TYPE_BAREMETAL_RHEL_SERVER:
        os_type = "RHELSERVER"
    elif node_type == NODE_TYPE_BAREMETAL_CENTOS_SERVER:
        os_type = "CENTOSSERVER"
    elif node_type == NODE_TYPE_BAREMETAL_SLES_SERVER:
        os_type = "SLESSERVER"
    elif node_type == NODE_TYPE_BAREMETAL_OEL_SERVER:
        os_type = "OELSERVER"
    elif node_type == NODE_TYPE_BAREMETAL_WINDOWS_SERVER:
        os_type = "WINDOWSSERVER"
    elif node_type == NODE_TYPE_BAREMETAL_RHEL_CONTAINER:
        os_type = "RHELCONTAINER"
    elif node_type == NODE_TYPE_BAREMETAL_CENTOS_CONTAINER:
        os_type = "CENTOSCONTAINER"
    else:
        # Should not happen
        os_type = "Other"
    api_ip_or_host, api_port = split_address(api_server)
    api_port = _MP_API_PORT if api_port is None else api_port
    mgmt_ips = get_management_ip_addresses(api_ip_or_host)
    os_version = get_os_version()
    fqdn = socket.getfqdn()
    reg_body = json.dumps({ "node": {
        "resource_type": "HostNode",
        "display_name": fqdn,
        "os_type": os_type,
        "os_version": os_version,
        "external_id" : node_uuid,
        "ip_addresses": mgmt_ips },
        "msg_client_info": {
            "account_name": account_name,
            "shared_secret": shared_secret,
            "certificate": host_cert
        }
    })
    try:
        status, json_resp, text, _ = nsx_api_request(api_ip_or_host, api_port,
                                                     api_thumbprint,
                                                     {"username": api_username,
                                                      "password": api_password,
                                                      "token": token},
                                                     reg_method,
                                                     reg_uri, body=reg_body,
                                                     headers=headers)
    except ApiException as ex:
        _restore_host_state(old_host_state)
        return MGR_REGISTER_EXCEPTION, None, get_message(ex.args[0])
    if status == compat23_httplib.UNAUTHORIZED:
        err_msg = "authentication failure"
        lg.info("Node registration authentication failure")
        _restore_host_state(old_host_state)
        return MGR_AUTHENTICATION_FAILED, None, err_msg
    elif status not in [compat23_httplib.CREATED, compat23_httplib.OK]:
        lg.info("Node registration failed, rc=%s: %s", status, text)
        if json_resp and "error_message" in json_resp:
            err_msg = json_resp["error_message"]
            lg.info("Error from MP API: %s", json_resp["error_message"])
        else:
            err_msg = ("the manager API server rejected the registration attempt")
        _restore_host_state(old_host_state)
        return MGR_REGISTER_FAILED, None, err_msg
    else:
        # compat23_httplib.CREATED or compat23_httplib.OK
        node_uuid = json_resp.get("id")
        headers = {"Content-Type": "application/json"}
        reg_method = "GET"
        reg_uri = "/api/v1/cluster/nodes/"
        reg_body = ""
        try:
            status, json_resp, text, _ = nsx_api_request(api_ip_or_host,
                                                         api_port,
                                                         api_thumbprint,
                                                         {"username": api_username,
                                                          "password": api_password,
                                                          "token": token},
                                                         reg_method,
                                                         reg_uri, body=reg_body,
                                                         headers=headers)
        except ApiException as ex:
            _restore_host_state(old_host_state)
            return MGR_REGISTER_EXCEPTION, None, get_message(ex.args[0])
        if status == compat23_httplib.UNAUTHORIZED:
            err_msg = "authentication failure when getting MP cluster"
            lg.info("authentication failure when getting MP cluster")
            _restore_host_state(old_host_state)
            return MGR_AUTHENTICATION_FAILED, None, err_msg
        elif status !=  compat23_httplib.OK:
            lg.info("Node registration failed, rc=%s: %s", status, text)
            err_msg = ("API error getting MP cluster")
            if "error_message" in json_resp:
                lg.info("Error getting MP cluster: %s", json_resp["error_message"])
            _restore_host_state(old_host_state)
            return MGR_REGISTER_FAILED, None, err_msg
        else:
            brokers = []
            aph_list = []
            for node in json_resp["results"]:
                if "manager_role" in node:
                    _b = node["manager_role"]["mgmt_plane_listen_addr"]
                    if ("certificate" in _b) and (("ip_address" in _b) or ("fqdn" in _b)) and ("port" in _b):
                        _broker = {
                            _MPA_BROKER_IP_ADDR_KEY: _b["ip_address"] if "ip_address" in _b else "",
                            _MPA_BROKER_FQDN_KEY: _b["fqdn"] if "fqdn" in _b else "",
                            _MPA_BROKER_PORT_KEY: _b["port"],
                            _MPA_BROKER_VIRTUAL_HOST_KEY: "nsx",
                            _MPA_BROKER_SSL_TP_KEY: sha256(ssl.PEM_cert_to_DER_cert(_b["certificate"])).hexdigest(),
                            _MPA_BROKER_IS_MASTER_KEY: "true"
                        }
                        brokers.append(_broker)
                    _aph_node = node["manager_role"].get("appliance_connection_info")
                    if _aph_node:
                        _aph = get_aph_info(_aph_node)
                        if _aph:
                            aph_list.append(_aph)

            # We need to restart both the old MPA service and nsx-proxy.
            # - nsx-proxy, because we need it to consume the new CLI gnenerated
            # cert and key.
            # - MPA, to disable it in case we received APH info and will be
            # switching to the TnProxy MPA, or if we received broker info.
            _aphconfig_update_services_action(False)
            _update_host_cert_key(host_cert, host_key)

            if aph_list:
                if not _write_appliance_info_xml(aph_list, restart_services=False):
                    lg.warning("_register: Failed to write appliance-info.xml")

            # TODO - Pick one of the MP cluster nodes as master. For now,
            # marking every node as master - we are only supporting
            # single-node MP cluster at this time
            if brokers:
                if _write_mpa_config(account_name, shared_secret, brokers, api_server,
                      api_thumbprint, token, node_id, restart_services=False) != 0:
                    # Log a warning, return SUCCESS.
                    # This should never happen since we checked
                    # above that the file exists and is writable
                    lg.warning("_register: Failed to write MPA config")
                _aphconfig_update_services_action(True)
                return MGR_SUCCESS, node_uuid, None
            else:
                err_msg = "Failed to get MP cluster information"

        _restore_host_state(old_host_state)
        _aphconfig_update_services_action(True)
        return MGR_REGISTER_FAILED, None, None


def _retrieve_uuid(api_server, api_username, api_thumbprint, api_password,
                   mgmt_ips):
    query_str = [ "ip_address=%s" % ip_addr for ip_addr in mgmt_ips ]
    reg_uri = "/nsxapi/api/v1/fabric/nodes?" + '&'.join(query_str)
    api_ip_or_host, api_port = split_address(api_server)
    api_port = _MP_API_PORT if api_port is None else api_port
    reg_method = "GET"
    headers = {"Content-Type": "application/json"}
    status, json_resp, text, _ = nsx_api_request(api_ip_or_host, api_port,
                                                 api_thumbprint,
                                                 {"username": api_username, "password": api_password},
                                                 reg_method,
                                                 reg_uri, headers=headers)
    return json_resp['results'][0]['id']


def _reset(api_server, api_username, api_thumbprint, api_password):
    """Recover node by invoking configured manager(s)' northbound API.
    Args:
        api_server: Hostname or IP address and optional :port of a MP API server
        api_thumbprint: Expected thumbprint of manager API server.
        api_username: Username to authenticate with MP API.
        api_password: Password to authenticate with MP API.
    Return:
        (MGR_SUCCESS, node_uuid, None) if registration succeeded.
        (MGR_AUTHENTICATION_FAILED, None, err_msg) tuple if authentication failed.
        (MGR_REGISTER_EXCEPTION, None, err_msg) tuple if exception is raised.
        (MGR_REGISTER_FAILED, None, err_msg) tuple in case of any other failure.
    """
    node_uuid = None
    account_name = None
    shared_secret = None
    if not os.path.isfile(FilePaths.MPA_CONFIG_FILE):
        try:
            open(FilePaths.MPA_CONFIG_FILE, 'w+').close()
        except:
            err_msg = "MPA config file is not writable"
            return MGR_REGISTER_FAILED, None, err_msg
    else:
        try:
            node_uuid, account_name, shared_secret = get_mpa_node_config()
        except ValueError:
            lg.warning("Ignoring corrupted MPA config file.")
    api_ip_or_host, api_port = split_address(api_server)
    if not node_uuid:
        try:
            mgmt_ips = get_management_ip_addresses(api_ip_or_host)
            node_uuid = _retrieve_uuid(api_server, api_username, api_thumbprint,
                                       api_password, mgmt_ips)
        except:
            pass
        if not node_uuid:
            return MGR_REGISTER_EXCEPTION, None, "Failed to retrieve node UUID"
        account_name = _HV_CLIENT_TYPE + "-" +  node_uuid

    host_key = ""
    host_cert = ""

    # Backup host state to restore in case of failure
    old_host_state = {}
    _backup_host_state(old_host_state)

    get_existing_cert = True
    needs_proxy_restart = True
    if _should_update_host_cert(node_uuid):
        # Create host cert/key pair. If this fails, then we revert to the old
        # behavior of nsx-proxy creating its own key/cert.
        ret, tmp_key_file, tmp_cert_file = _gen_host_cert(node_uuid)
        if ret:
            try:
                host_cert = open(tmp_cert_file, 'r').read()
                host_key  = open(tmp_key_file, 'r').read()
                get_existing_cert = False
            except Exception:
                lg.exception("Error reading temp cert/key",
                             error_code=ec.INTERNAL_ERROR)
            # We've read the cert/key file contents already. Clean them up now.
            for path in [tmp_key_file, tmp_cert_file]:
                try:
                    os.remove(path)
                except Exception:
                    lg.exception("Error cleaning up temp key/cert file: " + path,
                         error_code=ec.INTERNAL_ERROR)
            if not get_existing_cert:
                # Write host-cfg.xml, but do not restart nsx_proxy
                result = _write_host_cfg_xml(node_uuid, False)
                if not result:
                    err_msg = "Unable to update host-cfg.xml"
                    return MGR_REGISTER_FAILED, None, err_msg
    else:
        needs_proxy_restart = False

    if get_existing_cert:
        if needs_proxy_restart:
            # The resulting nsxproxy service restart updates certificates if
            # necessary.
            result = _write_host_cfg_xml(node_uuid, False)
            if not result:
                err_msg = "Unable to update host-cfg.xml"
                return MGR_REGISTER_FAILED, None, err_msg
        success, host_cert, host_key = _get_host_cert_key()
        if not success:
            lg.warning("_reset: Host cert not found at %s",
                       FilePaths.HV_CERT_FILE)

    if not shared_secret:
        shared_secret = _generate_account_secret()
    reg_method = "PUT"
    reg_uri = "/nsxapi/api/v1/messaging/clients/%s" % node_uuid
    api_port = _MP_API_PORT if api_port is None else api_port

    os_version = get_os_version()
    reg_body = json.dumps({
       "client_type" : "cvn-hv",
       "client_token" : node_uuid,
       "client_id" : node_uuid,
       "account_name" : account_name,
       "certificate": host_cert,
       "account_shared_secret": shared_secret
    })
    headers = {"Content-Type": "application/json"}
    try:
        status, json_resp, text, _ = nsx_api_request(api_ip_or_host, api_port,
                                                     api_thumbprint,
                                                     {"username": api_username, "password": api_password},
                                                     reg_method,
                                                     reg_uri, body=reg_body,
                                                     headers=headers)
    except ApiException as ex:
        _restore_host_state(old_host_state)
        return MGR_REGISTER_EXCEPTION, None, get_message(ex.args[0])
    if status == compat23_httplib.UNAUTHORIZED:
        err_msg = "authentication failure"
        lg.info("Node reset authentication failure")
        _restore_host_state(old_host_state)
        return MGR_AUTHENTICATION_FAILED, None, err_msg
    elif status not in [compat23_httplib.CREATED, compat23_httplib.OK]:
        lg.info("Node reset failed, rc=%s: %s", status, text)
        if json_resp and "error_message" in json_resp:
            err_msg = json_resp["error_message"]
            lg.info("Error from MP API: %s", json_resp["error_message"])
        else:
            err_msg = ("the manager API server rejected the registration attempt")
        _restore_host_state(old_host_state)
        return MGR_REGISTER_FAILED, None, err_msg
    else:
        # compat23_httplib.CREATED or compat23_httplib.OK
        headers = {"Content-Type": "application/json"}
        reg_method = "GET"
        reg_uri = "/nsxapi/api/v1/cluster/nodes/"
        reg_body = ""
        try:
            status, json_resp, text, _ = nsx_api_request(api_ip_or_host,
                                                         api_port,
                                                         api_thumbprint,
                                                         {"username": api_username, "password": api_password},
                                                         reg_method,
                                                         reg_uri, body=reg_body,
                                                         headers=headers)
        except ApiException as ex:
            _restore_host_state(old_host_state)
            return MGR_REGISTER_EXCEPTION, None, get_message(ex.args[0])
        if status == compat23_httplib.UNAUTHORIZED:
            err_msg = "authentication failure when getting MP cluster"
            lg.info("authentication failure when getting MP cluster")
            _restore_host_state(old_host_state)
            return MGR_AUTHENTICATION_FAILED, None, err_msg
        elif status !=  compat23_httplib.OK:
            lg.info("Node registration failed, rc=%s: %s", status, text)
            err_msg = ("API error getting MP cluster")
            if "error_message" in json_resp:
                lg.info("Error getting MP cluster: %s", json_resp["error_message"])
            _restore_host_state(old_host_state)
            return MGR_REGISTER_FAILED, None, err_msg
        else:
            brokers = []
            aph_list = []
            for node in json_resp["results"]:
                if "manager_role" in node:
                    _b = node["manager_role"]["mgmt_plane_listen_addr"]
                    if ("certificate" in _b) and (("ip_address" in _b) or ("fqdn" in _b)) and ("port" in _b):
                        _broker = {
                            _MPA_BROKER_IP_ADDR_KEY: _b["ip_address"] if "ip_address" in _b else "",
                            _MPA_BROKER_FQDN_KEY: _b["fqdn"] if "fqdn" in _b else "",
                            _MPA_BROKER_PORT_KEY: _b["port"],
                            _MPA_BROKER_VIRTUAL_HOST_KEY: "nsx",
                            _MPA_BROKER_SSL_TP_KEY: sha256(ssl.PEM_cert_to_DER_cert(_b["certificate"])).hexdigest(),
                            _MPA_BROKER_IS_MASTER_KEY: "true"
                        }
                        brokers.append(_broker)

                    _aph_node = node["manager_role"].get("appliance_connection_info")
                    if _aph_node:
                        _aph = get_aph_info(_aph_node)
                        if _aph:
                            aph_list.append(_aph)

            _aphconfig_update_services_action(False)
            _update_host_cert_key(host_cert, host_key)

            if aph_list:
                if not _write_appliance_info_xml(aph_list, restart_services=False):
                    # Log as warning, but don't fail. This behavior will be
                    # revisited in the near future.
                    lg.warning("_register: Failed to write appliance-info.xml")

            # TODO - Pick one of the MP cluster nodes as master. For now,
            # marking every node as master - we are only supporting
            # single-node MP cluster at this time
            if brokers:
                if _write_mpa_config(account_name, shared_secret, brokers, restart_services=False) != 0:
                    # Log a warning, return SUCCESS.
                    # This should never happen since we checked
                    # above that the file exists and is writable
                    lg.warning("_reset: Failed to write MPA config")
                _aphconfig_update_services_action(True)
                return MGR_SUCCESS, node_uuid, None
            else:
                err_msg = "Failed to get MP cluster information"

        _restore_host_state(old_host_state)
        _aphconfig_update_services_action(True)
        return MGR_REGISTER_FAILED, None, None


def _deregister(api_server, api_username, api_thumbprint, api_password):
    """Deregister node by invoking configured manager(s)' northbound API.
    Args:
        api_server: Hostname or IP address and optional :port of a MP API server
        api_thumbprint: Expected thumbprint of manager API server.
        api_username: Username to authenticate with MP API.
        api_password: Password to authenticate with MP API.
    Return:
        (MGR_SUCCESS, None) if deregistration succeeded.
        (MGR_AUTHENTICATION_FAILED, None) tuple if authentication failed.
        (MGR_DEREGISTER_EXCEPTION, err_msg) tuple if exception is raised.
        (MGR_DEREGISTER_FAILED, err_msg) tuple in case of any other failure.
    """
    if not os.access(FilePaths.MPA_CONFIG_FILE, os.W_OK):
        err_msg = "MPA config file does not exist or is not writable"
        return MGR_DEREGISTER_FAILED, err_msg
    node_uuid, account_name, shared_secret = get_mpa_node_config()
    if not node_uuid:
        err_msg = "Node UUID not found in MPA config file"
        lg.info("Node de-registration failed, UUID not found")
        return MGR_DEREGISTER_FAILED, err_msg
    api_ip_or_host, api_port = split_address(api_server)
    api_port = _MP_API_PORT if api_port is None else api_port
    reg_method = "DELETE"
    reg_body=""
    reg_uri = "/nsxapi/api/v1/fabric/nodes/" + node_uuid + "?unprepare_host=false"
    headers = {"Content-Type": "application/json"}
    try:
        status, json_resp, text, _ = nsx_api_request(api_ip_or_host, api_port,
                                                    api_thumbprint,
                                                    {"username": api_username, "password": api_password},
                                                    reg_method,
                                                    reg_uri, body=reg_body,
                                                    headers=headers)
    except ApiException as ex:
        return MGR_DEREGISTER_EXCEPTION, get_message(ex.args[0])
    if status == compat23_httplib.UNAUTHORIZED:
        err_msg = "authentication failure"
        lg.info("Node deregistration authentication failure")
        return MGR_AUTHENTICATION_FAILED, err_msg
    elif status != compat23_httplib.OK:
        lg.info("Node deregistration failed, rc=%s: %s", status, text)
        if json_resp and "error_message" in json_resp:
            err_msg = json_resp["error_message"]
            lg.info("Error from MP API: %s", json_resp["error_message"])
        else:
            err_msg = ("the manager API server rejected the deregistration attempt")
        return MGR_DEREGISTER_FAILED, err_msg
    else:
        # compat23_httplib.OK
        _aphconfig_update_services_action(False)
        if _write_mpa_config("", "",[], restart_services=False) != 0:
            # Log as warning, but return SUCCESS since we have
            # successfully deregistered node from MP,
            # but could not write empty config to mpaconfig file
            lg.warning("_deregister: Failed to write MPA config")
        if not _write_appliance_info_xml(restart_services=False):
            # Log as warning, but don't fail. This behavior will be
            # revisited in the near future.
            lg.warning("_deregister: Failed to delete appliance-info.xml")
        if not _write_host_cfg_xml('', False):
            lg.warning("_deregister: Failed to write host-cfg.xml")
        _aphconfig_update_services_action(True)
        return MGR_SUCCESS, None


class HostJoin(CliCommandProvider):
    def __call__(self, cmd_list):
        if is_reboot_required():
            return "MSG_INVALID_HOST_STATE_REBOOT_REQUIRED"
        username = self.get_arg_value("api-username", cmd_list)
        thumbprint = ApiThumbprintArgument.normalize(
            self.get_arg_value("api-thumbprint", cmd_list))
        token = self.get_arg_value("api-token", cmd_list)
        tokenfile = self.get_arg_value("file-name", cmd_list)
        node_uuid = self.get_arg_value("uuid-arg", cmd_list)
        # If password isn't entered by the user with the rest of the join
        # command, the framework prompts for it in which case the password
        # value is guaranteed to be the last argument value.
        if len(cmd_list) > 8:
            password = self.get_arg_value("api-password", cmd_list)
        else:
            password = cmd_list[-1]
        result, node_uuid, err_msg = _register(cmd_list[2], username,
                                               thumbprint, password,
                                               token, tokenfile, node_uuid)
        if result == MGR_AUTHENTICATION_FAILED:
            return "MSG_HOST_MANAGER_AUTHENTICATION_FAILED"
        elif result == MGR_REGISTER_EXCEPTION:
            return "MSG_HOST_REGISTER_EXCEPTION\n%s" % err_msg
        elif result == MGR_REGISTER_FAILED:
            err_msg = err_msg if err_msg else "internal error"
            return "MSG_HOST_MANAGER_REGISTRATION_FAILED\n%s" % err_msg
        return "OK\n\n%s" % json.dumps({"node_uuid": node_uuid})


class HostReset(CliCommandProvider):
    def __call__(self, cmd_list):
        username = self.get_arg_value("api-username", cmd_list)
        thumbprint = ApiThumbprintArgument.normalize(
            self.get_arg_value("api-thumbprint", cmd_list))
        # If password isn't entered by the user with the rest of the join
        # command, the framework prompts for it in which case the password
        # value is guaranteed to be the last argument value.
        if len(cmd_list) > 8:
            password = self.get_arg_value("api-password", cmd_list)
        else:
            password = cmd_list[-1]
        result, node_uuid, err_msg = _reset(cmd_list[2], username,
                                            thumbprint, password)
        if result == MGR_AUTHENTICATION_FAILED:
            return "MSG_HOST_MANAGER_AUTHENTICATION_FAILED"
        elif result == MGR_REGISTER_EXCEPTION:
            return "MSG_HOST_REGISTER_EXCEPTION\n%s" % err_msg
        elif result == MGR_REGISTER_FAILED:
            err_msg = err_msg if err_msg else "internal error"
            return "MSG_HOST_MANAGER_REGISTRATION_FAILED\n%s" % err_msg
        return "OK\n\n%s" % json.dumps({"node_uuid": node_uuid})


class HostDetach(CliCommandProvider):
    def __call__(self, cmd_list):
        username = self.get_arg_value("api-username", cmd_list)
        thumbprint = ApiThumbprintArgument.normalize(
            self.get_arg_value("api-thumbprint", cmd_list))
        # If password isn't entered by the user with the rest of the join
        # command, the framework prompts for it in which case the password
        # value is guaranteed to be the last argument value.
        if len(cmd_list) > 8:
            password = self.get_arg_value("api-password", cmd_list)
        else:
            password = cmd_list[-1]
        result, err_msg = _deregister(cmd_list[2], username,
                                      thumbprint, password)
        if result == MGR_AUTHENTICATION_FAILED:
            return "MSG_HOST_MANAGER_AUTHENTICATION_FAILED"
        elif result == MGR_DEREGISTER_EXCEPTION:
            return "MSG_HOST_DEREGISTER_EXCEPTION\n%s" % err_msg
        elif result == MGR_DEREGISTER_FAILED:
            err_msg = err_msg if err_msg else "internal error"
            return "MSG_HOST_MANAGER_DEREGISTRATION_FAILED\n%s" % err_msg
        return "OK\n\n%s" % json.dumps("")


# DEPRECATED: delete after flash release
class HostClear(CliCommandProvider):
    def __call__(self, cmd_list):
        #Copy script to tmp folder to execute externally as it will remove all vibs including nsxcli
        os.system("cp -p %s %s" % (FilePaths.RESET_NSX_SCRIPT, FilePaths.TEMP_RESET_NSX_SCRIPT))
        if is_windows():
            command_nsxa = [FilePaths.NSXA_PATH, "clear"]
        else:
            command_nsxa = [FilePaths.TEMP_RESET_NSX_SCRIPT, "fabric_node", node_type]
        rc, p1_out, p1_err = invoke_command(command_nsxa)
        if rc != 0:
            err_msg = ("Error in 'nsxa clear'. error code: %d"
                       " stdout: %s stderr: %s" %
                       (rc, p1_out, p1_err))
            lg.warning(err_msg);
            err_msg = err_msg.replace('\n',' ')
            return "MSG_HOST_CLEAR_EXCEPTION\n%s" % err_msg
        return "OK\n\n%s" % json.dumps("")


class HostDeleteNsx(CliCommandProvider):
    def __call__(self, cmd_list):
        if cmd_list[2].lower() in ["y", "yes"]:
            #Copy script to tmp folder to execute externally as it will remove all vibs including nsxcli
            os.system("cp -p %s %s" % (FilePaths.RESET_NSX_SCRIPT, FilePaths.TEMP_RESET_NSX_SCRIPT))
            if is_windows():
                command_nsxa = [FilePaths.NSXA_PATH, "clear"]
            else:
                command_nsxa = [FilePaths.TEMP_RESET_NSX_SCRIPT, "fabric_node", node_type]
            rc, p1_out, p1_err = invoke_command(command_nsxa)
            if rc != 0:
                err_msg = ("Error in 'nsx environment delete'. error code: %d"
                           " stdout: %s stderr: %s" %
                           (rc, p1_out, p1_err))
                err_msg = err_msg.replace('\n',' ')
                lg.warning(err_msg);
                if p1_out and 'Command to uninstall NSX :' in p1_out:
                    p1_out = "%s" % p1_out.split('Command to uninstall NSX :')[-1]
                err_msg = (" error code: %d"
                           " stdout: %s, stderr: %s" %
                           (rc, p1_out, p1_err))
                err_msg = err_msg.replace('\n',' ')
                return "MSG_HOST_DELETE_NSX_EXCEPTION\n%s" % err_msg
            return "OK\n\n%s" % json.dumps(_MSG_HOST_SUCCESS_DELETE_NSX)
        else:
            lg.info("User specified 'no' on 'del nsx' confirmation")
            return "OK\n\n%s" % json.dumps(_MSG_HOST_CANCEL_DELETE_NSX)


# DEPRECATED: delete after QE updates test cases to use new syntax
class HostRegister(CliCommandProvider):
    def __call__(self, cmd_list):
        result, node_uuid, err_msg = _register(cmd_list[1], cmd_list[2],
                                               cmd_list[3], cmd_list[4])
        if result == MGR_AUTHENTICATION_FAILED:
            return "MSG_HOST_MANAGER_AUTHENTICATION_FAILED"
        elif result == MGR_REGISTER_EXCEPTION:
            return "MSG_HOST_REGISTER_EXCEPTION\n%s" % err_msg
        elif result == MGR_REGISTER_FAILED:
            err_msg = err_msg if err_msg else "internal error"
            return "MSG_HOST_MANAGER_REGISTRATION_FAILED\n%s" % err_msg
        return "OK\n\n%s" % json.dumps({"node_uuid": node_uuid})


# DEPRECATED: delete after QE updates test cases to use new syntax
class HostDeregister(CliCommandProvider):
    def __call__(self, cmd_list):
        result, err_msg = _deregister(cmd_list[1], cmd_list[2],
                                      cmd_list[3], cmd_list[4])
        if result == MGR_AUTHENTICATION_FAILED:
            return "MSG_HOST_MANAGER_AUTHENTICATION_FAILED"
        elif result == MGR_DEREGISTER_EXCEPTION:
            return "MSG_HOST_DEREGISTER_EXCEPTION\n%s" % err_msg
        elif result == MGR_DEREGISTER_FAILED:
            err_msg = err_msg if err_msg else "internal error"
            return "MSG_HOST_MANAGER_DEREGISTRATION_FAILED\n%s" % err_msg
        return "OK\n\n%s" % json.dumps("")


class HostShowUUID(CliCommandProvider):
    def __call__(self, cmd_list):
        node_uuid, account_name, shared_secret = get_mpa_node_config()

        if node_uuid:
            return "OK\n\n%s" % json.dumps({"uuid": node_uuid})
        return "MSG_HOST_IS_NOT_REGISTERED"

