# ***************************************************************************
# Copyright 2017-2020 VMware, Inc.  All rights reserved. VMware Confidential.
# ***************************************************************************
import logging
from logging import CRITICAL
from logging import FATAL
from logging import ERROR
from logging import WARNING
from logging import WARN
from logging import INFO
from logging import DEBUG
from logging import NOTSET
from logging import raiseExceptions
from logging.handlers import SysLogHandler
from nsx_logging.config import fileConfig
from nsx_logging.formatter import StructuredDataFormatter
from nsx_logging.handlers import GetRotatingFileHandler


class Filter(object):
    """
    This is a customized class which is used to add filters.
    """
    def __init__(self, exactlist=None, startswith=None, contains=None, reverse=False):
        """
        exactlist will do exact matching,
        startswith will match log names that start with the key,
        contains will match log names that contain the key,
        reverse will reverse the matching.
        The priority will be exactlist > startswith > contains.
        """
        self.exactlist =\
            exactlist.split() if isinstance(exactlist, str) else exactlist
        self.startswith =\
            startswith.split() if isinstance(startswith, str) else startswith
        self.contains =\
            contains.split() if isinstance(contains, str) else contains
        self.reverse = reverse

    def filter(self, record):
        result = False
        if isinstance(self.exactlist, list):
            if record.name in self.exactlist:
                result = True
        elif isinstance(self.startswith, list):
            for key in self.startswith:
                if record.name.startswith(key):
                    result = True
                    break
        elif isinstance(self.contains, list):
            for key in self.contains:
                if key in record.name:
                    result = True
                    break
        else:
            pass
        return not result if self.reverse else result


class NsxLogger(logging.Logger):
    def getStructuredData(self, kwargs):
        """
        The original logging.Logger will only process field named 'extra'.
        This method will be used to insert every field in white list into the
        'extra' dictionary.
        """
        structure_data = {}
        for k, v in list(kwargs.items()):
            structure_data[self._format(k)] = v
        return structure_data

    def _format(self, field):
        # convert Ptyhon structured data fields to camel case
        # for example, error_code will be converted to errorCode
        tokens = field.split("_")
        for i in range(1, len(tokens)):
            tokens[i] = tokens[i].capitalize()
        return "".join(tokens)

    def _log(self, level, msg, args, exc_info=None, extra=None, **kwargs):
        structured_data = self.getStructuredData(kwargs)
        # Lower level to WARNING if error code is not set
        # Ignore ERROR and CRITICAL logs from the alarm framework
        if level in [ERROR, CRITICAL] and "errorCode" not in structured_data\
                and "eventType" not in structured_data:
            level = WARNING
            structured_data["invalid"] = "true"
        if len(structured_data) != 0 and extra is None:
            extra = {"tags": structured_data}
        super(NsxLogger, self)._log(level, msg, args, exc_info, extra)


def getLogger(name=None):
    """
    This method will override the Logger class as in 'logging.Logger', which
    will override some Logger behavior and thus support structured data
    white-list.
    """
    lg = logging.getLogger(name)
    lg.__class__ = NsxLogger
    return lg


def _get_nsx_node_type():
    """
    This is a private helper method which reads '/etc/nsx_issue' and determine
    current node type.
    """
    try:
        with open("/etc/nsx_issue") as fd:
            lines = fd.readlines()
        attribs = {}
        for line in lines:
            parts = line.split(":")
            if len(parts) == 2 and parts[0].strip() == "node-type":
                node_types = parts[1].strip().split(" ")
                # Use nsx-manager if exists, otherwise use the first node type
                if "nsx-manager" in node_types:
                    return "nsx-manager"
                return node_types[0]
    except:
        pass
    # default node type is 'xxx', indicating an error
    return "xxx"


# this api is marked as unstable.
# avoid using this for now.
def basicConfig(syslog=False, stdout=False, fmt=None, comp=None, subcomp=None,
                msgid="-", ident="nsx_logger", *args, **kwargs):
    """
    This method is a light-weight configurator, which can make logger logs
    to a file (and syslog).
    Notice: syslog will always use RFC5424. The fmt is just for file and stdout.
    """
    lg = getLogger()
    lg.handlers = []
    node_type = _get_nsx_node_type()
    comp = comp or node_type
    subcomp = subcomp or "python"
    custom_fmt = rfc5424 = StructuredDataFormatter(comp=comp, subcomp=subcomp,
        msgid=msgid, node_type=node_type)
    if syslog is True:
        if node_type == "nsx-esx":
            syslog_hdl = SysLogHandler(
                address="/dev/log",
                facility=SysLogHandler.LOG_LOCAL6)
            # the following colon is necessary for the ident to be recognized
            # by vmsyslogd on ESX hosts
            syslog_hdl.ident = ident + ":"
        else:
            syslog_hdl = SysLogHandler(
                address=("127.0.0.1", 514),
                facility=SysLogHandler.LOG_LOCAL6)
        syslog_hdl.setFormatter(rfc5424)
        lg.addHandler(syslog_hdl)
    if fmt:
        custom_fmt = logging.Formatter(fmt)
    file_hdl = GetRotatingFileHandler(**kwargs)
    if file_hdl is not None:
        file_hdl.setFormatter(custom_fmt)
        lg.addHandler(file_hdl)
    if stdout:
        import sys
        stdout_hdl = logging.StreamHandler(sys.stdout)
        stdout_hdl.setFormatter(custom_fmt)
        lg.addHandler(stdout_hdl)


def init(filename):
    """
    Load a logging config file in yaml format.
    Use logging.config.dictConfig to load the configurations.
    """
    fileConfig(filename)

def _defaultConfig():
    try:
        lg = getLogger()
        rfc5424 = StructuredDataFormatter(comp=_get_nsx_node_type(),
                                          subcomp="python")
        syslog_hdl = SysLogHandler(facility=SysLogHandler.LOG_LOCAL6)
        syslog_hdl.setFormatter(rfc5424)
        lg.addHandler(syslog_hdl)
    except:
        pass

# By default, log will go to syslog
# This setting will be executed only once, when this library is first loaded.
# Will be overwritten if user specifies a configuration.
_defaultConfig()
