# Copyright (C) 2020 VMware, Inc. All rights reserved.
# -- VMware Confidential

import sha.config
import ast
import traceback

from sha.contrib.metric.abc_base import Metric
from sha.contrib.metric.utils import run_command, timeout_cache


class HostMetrics(Metric):
    def __init__(self):
        self._logger = sha.config.get_logger(__name__)

    def get_vmnic_device_stats(self, vmnic):
        device_stats_cmd = ['vsish', '-p', '-e', 'get', 'net/pNics/{}/stats'.format(vmnic)]
        try:
            output = run_command(device_stats_cmd, shell=False) or None
        except Exception as e:
            trace = traceback.format_exc()
            self._logger.warning("Failed to get pNic device stats {}:{}".format(e, trace))
            output = None
        if not output:
            return None

        output = ast.literal_eval(output.replace('\n', ' '))
        ans = {
            'txpkts': output.get('txpkt', 0),
            'txbytes': output.get('txbytes', 0),
            'txerrs': output.get('txerr', 0) + output.get('txdrp', 0) + output.get('txaborterr', 0) +\
                      output.get('txcarerr', 0) + output.get('txfifoerr', 0) + output.get('txhearterr', 0) +\
                      output.get('txwinerr', 0),
            'rxpkts': output.get('rxpkt', 0),
            'rxbytes': output.get('rxbytes', 0),
            'rxerrs': output.get('rxerr', 0) + output.get('rxdrp', 0) + output.get('rxlgterr', 0) +\
                      output.get('rxoverr', 0) + output.get('rxcrcerr', 0) + output.get('rxfrmerr', 0) +\
                      output.get('rxfifoerr', 0) + output.get('rxmisserr', 0)
        }
        return ans

    def get_vmnic_stats(self, vmnic):
        cmd1 = ['net-stats', '-l']
        try:
            output = run_command(cmd1, shell=False) or None
        except Exception as e:
            trace = traceback.format_exc()
            self._logger.warning("Failed to get net-stats result {}:{}".format(e, trace))
            output = None
        if not output:
            return None

        vmnic0 = next((x.split() for x in output.splitlines() if vmnic in x), None)
        if not vmnic0:
            return None

        port, switch = vmnic0[0], vmnic0[3]
        cmd2 = ['vsish', '-p', '-e', 'get', 'net/portsets/{}/ports/{}/clientStats'.format(switch, port)]
        try:
            output = run_command(cmd2, shell=False) or None
        except Exception as e:
            trace = traceback.format_exc()
            self._logger.warning("Failed to get pNic switch stats {}:{}".format(e, trace))
            output = None
        if not output:
            return None

        output = ast.literal_eval(output)
        ans = {
            'txpkts': output.get('pktsTxOK', 0) + output.get('pktsTsoTxOK', 0),
            'txbytes': output.get('bytesTxOK', 0) + output.get('bytesTsoTxOK', 0),
            'txerrs': output.get('droppedTx', 0),
            'rxpkts': output.get('pktsRxOK', 0),
            'rxbytes': output.get('bytesRxOK', 0),
            'rxerrs': output.get('droppedRx', 0)
        }

        return ans

    def get_vmnic_sched_world_stats(self, vmnic):
        ps_cmd = ['ps']
        try:
            ps_output = run_command(ps_cmd, shell=False) or "None"
        except Exception as e:
            trace = traceback.format_exc()
            self._logger.warning("Failed to get world ID {}:{}".format(e, trace))
            ps_output = None
        if ps_output is None:
            return

        list_cmd = ['vsish', '-p', '-e', 'ls', 'sched/Vcpus']
        try:
            list_output = run_command(list_cmd, shell=False) or None
        except Exception as e:
            trace = traceback.format_exc()
            self._logger.warning("Failed to get vcpu list {}:{}".format(e, trace))
            list_output = None
        if list_output is None:
            return

        ans = {}
        hclk_sched = 'hclk-sched-{}'.format(vmnic)
        pollWorld = '{}-pollWorld'.format(vmnic)
        netWorld = 'NetWorld'
        vcpus = [vcpu.split() for vcpu in ps_output.splitlines() if vcpu != '' and \
                 (hclk_sched in vcpu or pollWorld in vcpu or netWorld in vcpu)]
        for vcpu in vcpus:
            self._logger.debug("vcpu is {}".format(vcpu[0]))
            if vcpu[0] not in list_output:
                continue
            name_cmd = ['vsish', '-p', '-e', 'get', 'sched/Vcpus/{}/worldName'.format(vcpu[0])]
            try:
                name_output = run_command(name_cmd, shell=False) or None
            except Exception as e:
                trace = traceback.format_exc()
                self._logger.warning("Failed to get world name {}:{}".format(e, trace))
                name_output = None
            if name_output is None:
                continue

            name = name_output.strip('\n').strip('"')
            stats_cmd = ['vsish', '-p', '-e', 'get', 'sched/Vcpus/{}/stats/stateTimes'.format(vcpu[0])]
            try:
                stats_output = run_command(stats_cmd, shell=False) or None
            except Exception as e:
                trace = traceback.format_exc()
                self._logger.warning("Failed to get world stats{}:{}".format(e, trace))
                stats_output = None
            if stats_output is None:
                continue

            stats_str = ast.literal_eval(stats_output.replace('\n', ' '))
            self._logger.debug("Vcpu {}'s stats is {}".format(name, stats_str))
            stats = {"used":stats_str.get("runTime", 0), "ready":stats_str.get("readyTime", 0)}
            self._logger.debug("stats is {}".format(stats))
            ans[name] = stats

        return ans

    def get_vmnic_sched_pktdrop_stats(self, vmnic):
        netstats_cmd = ['net-stats', '-l']
        try:
            netstats_output = run_command(netstats_cmd, shell=False) or None
        except Exception as e:
            trace = traceback.format_exc()
            self._logger.warning("Failed to get net-stats result {}:{}".format(e, trace))
            netstats_output = None
        if netstats_output is None:
            return

        ls_cmd = ['vsish', '-p', '-e', 'ls', '/vmkModules/netsched_hclk/devs/']
        try:
            ls_output = run_command(ls_cmd, shell=False) or None
        except Exception as e:
            trace = traceback.format_exc()
            self._logger.warning("Failed to get vm list {}:{}".format(e, trace))
            ls_output = None
        if ls_output is None or vmnic not in ls_output:
            return

        ans = {}
        ports = [port.split() for port in netstats_output.splitlines() if port != '' and "eth1" in port]
        for port in ports:
            self._logger.debug("port is {}".format(port))
            ls_port_cmd = ['vsish', '-p', '-e', 'ls', '/vmkModules/netsched_hclk/devs/{}/qleaves'.format(vmnic)]
            try:
                ls_port_output = run_command(ls_port_cmd, shell=False) or None
            except Exception as e:
                trace = traceback.format_exc()
                self._logger.warning("Failed to get vm {}:{}".format(e, trace))
                ls_port_output = None
            if ls_port_output is None or port[0] not in ls_port_output:
                continue

            self._logger.debug("sw port is {}".format(port[0]))
            netsched_cmd = ['vsish', '-p', '-e', 'get',
                            '/vmkModules/netsched_hclk/devs/{}/qleaves/netsched.pools.vm.{}/info'.format(vmnic, port[0])]
            try:
                output = run_command(netsched_cmd, shell=False) or None
            except Exception as e:
                trace = traceback.format_exc()
                self._logger.warning("Failed to get sw port pkt stats {}:{}".format(e, trace))
                output = None
            if output is None:
                continue

            name = "VM-{}".format(port[0])
            self._logger.debug("The stats for sw port {} is {}".format(port[0], output))
            stats = {'pktsDropped': ast.literal_eval(output).get('pktsDropped', 0)}
            self._logger.debug("stats list is {}".format(stats))
            ans[name] = stats

        return ans

    def get_vmnic_ena_stats(self, vmnic):
        # output:
        # ...
        # ===
        # bw_in_allowance_exceeded: 0
        # bw_out_allowance_exceeded: 0
        # pps_allowance_exceeded: 0
        # conntrack_allowance_exceeded: 0
        # linklocal_allowance_exceeded: 0

        ena_stats_cmd = ['localcli', '--plugin-dir', '/usr/lib/vmware/esxcli/int',
                         'networkinternal', 'nic', 'privstats', 'get', '-n',
                         '{}'.format(vmnic)]
        try:
            output = run_command(ena_stats_cmd, shell=False) or None
        except Exception as e:
            trace = traceback.format_exc()
            self._logger.warning("Failed to get pNic privstats {}:{}".format(e, trace))
            output = None
        if not output:
            return None

        ans = {}
        ena_keys = ['bw_in_allowance_exceeded', 'bw_out_allowance_exceeded', 'pps_allowance_exceeded',
                    'conntrack_allowance_exceeded', 'linklocal_allowance_exceeded']
        for line in output.splitlines():
            if (":" in line) and (line.split(":")[0] in ena_keys) :
                ans[line.split(":")[0]] = int(line.split(":")[1])
        return ans

    def get_host_metrics(self):
        ans = {}
        vmnic0_stats = self.get_vmnic_stats("vmnic0")
        if vmnic0_stats:
            ans['vmnic0.switch.stats'] = vmnic0_stats

        vmnic0_device_stats = self.get_vmnic_device_stats("vmnic0")
        if vmnic0_device_stats:
            ans['vmnic0.device.stats'] = vmnic0_device_stats

        vmnic0_sched_stats = self.get_vmnic_sched_world_stats("vmnic0")
        if vmnic0_sched_stats:
            ans['vmnic0.sched.stats'] = vmnic0_sched_stats

        vmnic0_sched_dropPkts = self.get_vmnic_sched_pktdrop_stats("vmnic0")
        if vmnic0_sched_dropPkts:
            ans['vmnic0.sched.dropPkts'] = vmnic0_sched_dropPkts

        vmnic0_ena_stats = self.get_vmnic_ena_stats("vmnic0")
        if vmnic0_ena_stats:
            ans['vmnic0.ena'] = vmnic0_ena_stats

        return ans

    @property
    def wavefront_source(self):
        # Get hostname as source to export metrics to wavefront
        return sha.config.HOST_NAME.lower()

    @timeout_cache(20)
    def usage(self, *args, **kwargs):
        return self.get_host_metrics(), self.wavefront_source
