# Copyright (C) 2020 VMware, Inc. All rights reserved.
# -- VMware Confidential

import struct
import traceback
import sha.config
from vmware.nsx.sha import sha_pm_pb2
from sha.contrib.monitor.processStatus.internal.client import Client

class ClientStopException(Exception):
    pass

class Parser(object):
    """
    Receive and parse the clients' message
    """
    def __init__(self, sock, client):
        self._logger = sha.config.get_logger(__name__)
        self._conn = sock
        self._client = client

    def _decode_msg(self, data):
        msg = sha_pm_pb2.ShaClientMsg()
        msg.ParseFromString(data)
        ns = msg.name_space
        self._logger.debug("PSM:name space is {}".format(ns))
        data = []
        for item in msg.item:
            self._logger.debug("PSM:data key is {} {}".format(item.key, item.type))
            if item.type == sha_pm_pb2.ItemMsgType.Value('NUMERICAL'):
                data.append({'key': item.key, 'val': item.numerical_value, 'ts': item.timestamp})
            elif item.type == sha_pm_pb2.ItemMsgType.Value('BOOLEAN'):
                data.append({'key': item.key, 'val': item.boolean_value, 'ts': item.timestamp})
            elif item.type == sha_pm_pb2.ItemMsgType.Value('STRING'):
                data.append({'key': item.key, 'val': item.string_value, 'ts': item.timestamp})
            else:
                data.append({'key': item.key, 'ts': item.timestamp})
        self._logger.debug("PSM:parsed data list is {}".format(data))

        return ns, data

    def _handle_data(self, ns, data):
        for item in data:
            if 'start' in item['key']:
                Client.add_client(self._client, ns)
            elif 'stop' in item['key']:
                Client.del_client(ns)
                raise ClientStopException
            elif 'heartbeat' in item['key']:
                self._client.update_heartbeat_ts(item['ts'])
            else:
                self._client.refresh_data(item)

    def recv(self):
        """Anything wrong with reception would cause the reconnection."""
        delimiter = self._conn.recv(9)
        self._logger.debug("PSM:delimiter is {}".format(delimiter.decode("utf-8")))
        self._conn.send("bumping".encode("utf-8"))
        if 'SHACLIENT' not in delimiter.decode("utf-8"):
            return
        (length,) = struct.unpack('h', self._conn.recv(2))
        self._logger.debug("PSM:payload length is {}".format(length))
        if length:
            bytes = self._conn.recv(length)
        else:
            raise Exception
        ns, data = self._decode_msg(bytes)
        self._handle_data(ns, data)

