# Copyright (C) 2019 VMware, Inc. All rights reserved.
# -- VMware Confidential

import queue
import threading
import traceback
import time

import sha.config
from sha.contrib.monitor.abc_base import Monitor
from sha.contrib.monitor.uplinkstatus.esx.connection import KcpConnection
from sha.contrib.monitor.uplinkstatus.esx.notification_handle import handle_notification

from sha.contrib.monitor.uplinkstatus.esx.message import InitNotification


class EsxUplinkStatusMonitor(Monitor):
    """
    EsxUplinkStatusMonitor monitors the uplink state changes happened on
    ESX host. The uplink status changes include:
    - Uplink state up/down
    - LAG member link state up/down
    """

    def __init__(self):
        self._logger = sha.config.get_logger(__name__)
        self._stop = False
        self._notification_queue = queue.Queue()
        self._msg_listen_thread = threading.Thread(
            name="nsx-sha-esx-uplink-monitor-msg-listen", target=self._listen_msg, daemon=True)
        self._msg_handle_thread = threading.Thread(
            name="nsx-sha-esx-uplink-monitor-msg-handle", target=self._handle_notification, daemon=True)
        self._kcp_connection = KcpConnection()
        self._status_cache = dict()

    def start(self):
        self._logger.info("Starting EsxUplinkStatusMonitor.")
        # 1. Establish channel to KCP_SHA kernel module
        # 2. Send init notification to get the init status
        # 3. Listen msgs and dispatch notification
        while not self._stop:
            try:
                if not self._kcp_connection.connect_to_kcp():
                    self._logger.error("Connect to kcp module failed")
                    time.sleep(10)
                    continue
                self._notification_queue.put(InitNotification())
                self._msg_handle_thread.start()
                self._msg_listen_thread.start()
                self._logger.info("EsxUplinkStatusMonitor started.")
                return
            except Exception as e:
                trace = traceback.format_exc()
                self._logger.error("Error in starting EsxUplinkStatusMonitor, Exception:{}, trace:{}".format(e, trace))
                time.sleep(10)

    def _listen_msg(self):
        while not self._stop:
            try:
                notification = self._kcp_connection.recv_notification()
                self._logger.debug("Received kcp notification:{}".format(notification))
                self._notification_queue.put(notification)
            except Exception as e:
                trace = traceback.format_exc()
                self._logger.error("Error in _listen_msg, Exception:{}, trace:{}".format(e, trace))
                time.sleep(1)
        self._logger.info("_msg_listen_thread stopped.")

    def _handle_notification(self):
        while not self._stop:
            try:
                notification = self._notification_queue.get(timeout=1)
                status_changes = dict()
                handle_notification(notification, status_changes)
                for alarm, status in status_changes.items():
                    if status != self._status_cache.get(alarm, None):
                        self._status_cache[alarm] = status
                        self.notify({alarm: status})
                self._notification_queue.task_done()
            except queue.Empty:
                continue
            except Exception as e:
                trace = traceback.format_exc()
                self._logger.error("Process UplinkStatus notification failed, Exception:{}, trace:{}".format(e, trace))
        self._logger.info("_msg_handle_thread stopped.")

    def stop(self):
        self._logger.info("Stopping EsxUplinkStatusMonitor.")
        self._stop = True
        try:
            self._kcp_connection.close()
        except Exception as e:
            trace = traceback.format_exc()
            self._logger.error("Error in closing kcp_connection, Exception:{}, trace:{}".format(e, trace))
        self._logger.info("EsxUplinkStatusMonitor is stopped.")
