# Copyright (C) 2019 VMware, Inc. All rights reserved.
# -- VMware Confidential
import datetime
import importlib.util
import sys
from keyword import iskeyword
from types import ModuleType
import sha.config
_logger = sha.config.get_logger(__name__)

from RestrictedPython import Guards, PrintCollector, safe_builtins
from RestrictedPython.Eval import default_guarded_getitem, default_guarded_getattr, default_guarded_getiter


class RestrictedModuleLoader:

    """Dynamically load modules for different restricted plugins"""

    _forbidden_modules = {'os', 'subprocess', 'sys', 'multiprocessing', 'importlib'}

    # permitted interface for sys module
    safe_for_sys = {
        'sys_path': sys.path,
        'sys_getsizeof': sys.getsizeof,
        'sys_getrefcount': sys.getrefcount,
        'sys_maxsize': sys.maxsize,
        'sys_getrecursionlimit': sys.getrecursionlimit,
    }

    def __init__(self, mpa_hub, alarm_hub, wf_hub):
        self._mpa_hub = mpa_hub
        self._alarm_hub = alarm_hub
        self._wf_hub = wf_hub
        # Bypass RestrictedPython validation in Guards.py, line 268, 'Using format() on a %s is not safe'
        safe_builtins['_getattr_'] = getattr

        # substitute interface for datetime module to avoid KeyError '__import__'
        substitute_interface_for_datetime = {
            'datetime_date_today': self.datetime_date_today,
            'datetime_date_strftime': self.datetime_date_strftime,
        }

        self._default_safe_globals = {
            '__builtins__': safe_builtins.copy(),
            '__metaclass__': type,
            '_iter_unpack_sequence_': Guards.guarded_iter_unpack_sequence,
            '_unpack_sequence_': Guards.guarded_unpack_sequence,
            '_write_': Guards.full_write_guard,
            '_getitem_': default_guarded_getitem,
            '_getiter_': default_guarded_getiter,
            '_getattr_': default_guarded_getattr,
            '_inplacevar_': self.custom_inplacevar,
            '_print_': PrintCollector,
            'all': all,
            'any': any,
            'map': map,
            'set': set,
            'dict': dict,
            'type': type,
            'list': list,
            **safe_builtins,
            **self.generate_sha_module().__dict__,
            **self.safe_for_sys,
            **substitute_interface_for_datetime
        }

    def _generate_action_modules(self):
        import importlib
        actions = {}
        for name in sha.config.ENABLED_ACTIONS:
            action_module = importlib.import_module('sha.contrib.action.{}'.format(name.lower()))
            action_clazz = getattr(action_module, name)
            if name == 'Alarm':
                actions[name] = getattr(action_clazz, 'do')(self._alarm_hub)
            elif name == 'Exportdata':
                actions[name] = getattr(action_clazz, 'do')(self._wf_hub)
            else:
                actions[name] = getattr(action_clazz, 'do')(self._mpa_hub)

        return actions

    def generate_sha_module(self):
        sha_module = ModuleType('sha')
        for name, action in self._generate_action_modules().items():
            setattr(sha_module, name, action)

        return sha_module

    def _allow_module(self, module_name, safe_globals):
        base_name = self.get_base_module_name(module_name)
        # Check if the module is already loaded before
        module_obj = importlib.import_module(module_name)
        base_module_obj = importlib.import_module(base_name)
        _logger.debug(
            "Load via import, module_name:{}, module:{}, base_name:{}, base_module:{}"
            .format(module_name, module_obj, base_name, base_module_obj))

        safe_globals[base_name] = base_module_obj
        safe_globals[module_name] = module_obj
        short_name = self.get_module_short_name(module_name)
        safe_globals[short_name] = module_obj

    def _allow_class(self, class_name, safe_globals):
        module_name, short_cls_name = self.parse_class_name(class_name)
        module = importlib.import_module(module_name)
        class_obj = getattr(module, short_cls_name)
        safe_globals[short_cls_name] = class_obj

    def get_safe_globals(self, name, module_list=None, class_list=None):
        """Plugins need to explicitly announce the modules they will use in code.
        The module name in the list may be a simple or dotted module or package
        name."""

        if class_list is None:
            class_list = []
        if module_list is None:
            module_list = []
        safe_globals = {}
        for module in module_list:
            if not self.is_valid_module_name(module):
                _logger.error("Invalid module name: {}".format(module))
                raise SyntaxError("Invalid module name: {}".format(module))

            if self.is_forbidden_module(module):
                _logger.error("{} is not allowed to use module: {}".format(name, module))
                raise ValueError('Module:{} is forbidden to use. Forbidden modules:{}'
                                 .format(module, list(self.forbidden_modules())))
            self._allow_module(module, safe_globals)

        for class_name in class_list:
            if not self.is_valid_class_name(class_name):
                _logger.error("Invalid class name: {}".format(class_name))
                raise SyntaxError("Invalid class name: {}".format(class_name))

            if self.is_class_from_forbidden_module(class_name):
                _logger.error("{} is not allowed to use class: {}".format(name, class_name))
                raise ValueError('Module of class:{} is forbidden to use. Forbidden modules:{}'
                                 .format(class_name, list(self.forbidden_modules())))

            self._allow_class(class_name, safe_globals)

        safe_globals.update(self._default_safe_globals)
        return safe_globals

    @staticmethod
    def custom_inplacevar(op, x, y):
        globs = {'x': x, 'y': y}
        exec('x' + op + 'y', globs)
        return globs['x']

    @classmethod
    def forbidden_modules(cls):
        return cls._forbidden_modules

    @staticmethod
    def get_module_short_name(module_name):
        dot = module_name.rfind('.')
        if dot > 0:
            return module_name[(dot + 1):]
        else:
            return module_name

    @staticmethod
    def get_base_module_name(module_name):
        dot = module_name.find('.')
        if dot > 0:
            return module_name[:dot]
        else:
            return module_name

    @staticmethod
    def datetime_date_today():
        """
        substitute for datetime.date.today()
        """
        return datetime.date.today()

    @staticmethod
    def datetime_date_strftime(dt, fmt):
        """
        substitute for datetime.date.strftime(self, fmt)
        :param dt: date instance
        :param fmt: format string
        """
        return dt.strftime(fmt)

    @classmethod
    def is_class_from_forbidden_module(cls, class_name):
        mod_name = cls.parse_class_name(class_name)[0]
        if cls.is_forbidden_module(mod_name):
            _logger.error("{} is not allowed to use.".format(mod_name))
            return True
        return False

    @classmethod
    def is_forbidden_module(cls, mod_name):
        base_mod_name = cls.get_base_module_name(mod_name)
        if mod_name in cls._forbidden_modules or base_mod_name in cls._forbidden_modules:
            _logger.error("{} is not allowed to use.".format(mod_name))
            return True
        return False

    @classmethod
    def parse_class_name(cls, class_name):
        """
        convert class name to module_name and short_class_name.
        :param class_name (e.g. 'datetime.date')
        :return: module_name, short_class_name (e.g. 'datetime', 'date')
        """
        dot = class_name.rfind('.')
        return class_name[:dot], class_name[(dot + 1):]

    @classmethod
    def is_valid_class_name(cls, class_name):
        """
        valid class name should be dot-separated legal name
        And should contain at least one dot
        """
        dot = class_name.rfind('.')
        if dot <= 0 or dot == len(class_name) - 1:
            return False

        mod_name = class_name[:dot]
        cls_short_name = class_name[(dot + 1):]
        if not cls.is_valid_module_name(mod_name):
            return False

        return cls.is_legal_name(cls_short_name)

    @classmethod
    def is_valid_module_name(cls, module_name):
        """
        valid module name should be dot-separated legal name
        """
        ms = module_name.split('.')

        for m in ms:
            if not cls.is_legal_name(m):
                return False

        return True

    @staticmethod
    def is_legal_name(name):
        """
        should be identifier and not keyword
        """
        return name.isidentifier() and not iskeyword(name)

