# Copyright (C) 2019-2020 VMware, Inc. All rights reserved.
# -- VMware Confidential
import queue
import threading
import traceback
from sha import config


class Hub(threading.Thread):
    """Dispatch msg to provider"""

    def __init__(self, provider, name, *args, **kwargs):
        super().__init__(name=name, *args, **kwargs)
        self._logger = config.get_logger(__name__)
        self._queue = queue.Queue(500)
        self._is_running = False
        self._provider = provider

    def __repr__(self):
        return self.name

    def send(self, msg):
        self._logger.debug("Sending msg:{} to hub:{}.".format(msg, self.name))
        try:
            self._queue.put_nowait(msg)
            self._logger.debug("Finished in sending msg:{} to hub:{}, Queue size:{}."
                               .format(msg, self.name, self._queue.qsize()))
        except queue.Full:
            self._logger.error("{} queue is full. Queue size:{}. Failed in sending msg:{} to hub. "
                               .format(self.name, self._queue.qsize(), msg))
        except Exception:
            trace = traceback.format_exc()
            self._logger.error("Failed in sending msg:{} to hub :{}. Queue size:{}. trace:{}"
                               .format(msg, self.name, self._queue.qsize(), trace))

    def run(self):
        self._logger.info('Running hub:{}'.format(self.name))
        self._is_running = True
        while self._is_running or not self._queue.empty():
            msg = None
            try:
                msg = self._queue.get(timeout=1)
                self._queue.task_done()
                self._logger.debug('{} is dispatching msg: {} to provider'
                                   .format(self.name, msg))
                try:
                    self._provider.sync(msg)
                    self._logger.debug('{} finished dispatching msg to provider, msg queue size:{}'
                                       .format(self.name, self._queue.qsize()))
                except Exception:
                    trace = traceback.format_exc()
                    self._logger.warning("{} failed in dispatching msg:{} to provider, trace:{}".
                                         format(self.name, msg, trace))
            except queue.Empty:
                pass
            except Exception:
                trace = traceback.format_exc()
                self._logger.error("Hub:{} - failed in handling msg: {}, trace: {}".format(self.name, msg, trace))

        self._provider.close()
        self._logger.info("stopped hub:{}".format(self.name))

    def stop(self):
        self._logger.info("going to stop hub:{}".format(self.name))
        self._is_running = False



