# Copyright (C) 2019-2020 VMware, Inc. All rights reserved.
# -- VMware Confidential
import os
import yaml
import traceback
from RestrictedPython import compile_restricted
from sha.core.accesscontrol import RestrictedModuleLoader
import uuid
from sha.core.plugin.plugin_logger import PluginLogger
from sha.core.profile._resolver import Resolver
from sha.contrib.metric.utils import run_command
from sha.constants import PRE_CONDITION_STR, ID, METRICS_STR, REACHED, UNREACHED_MSG, MODULES, CLASSES, DELAY_BEFORE_FIRST_CHECK
from sha.config import PRE_CONDITION, ENABLED_METRICS, PRE_INSTALL_SCRIPTS
from sha import config
from sha.core.channel.hub_holder import HubHolder
from sha.config_monitor import ConfigMonitor


class Plugin:
    _logger = config.get_logger(__name__)

    def __init__(self, id_, is_custom_plugin, safe_globals, compiled_script, name, manifest):
        self._id = id_
        self._is_custom_plugin = is_custom_plugin
        self._safe_globals = safe_globals
        self._compiled_script = compiled_script
        self._name = name
        self._metric_usage = []
        self._profile = None
        self._pre_conditions = manifest.get(PRE_CONDITION_STR, [])
        self._delay_before_first_check = manifest[DELAY_BEFORE_FIRST_CHECK] \
            if DELAY_BEFORE_FIRST_CHECK in manifest else 0
        self._manifest = manifest
        self._metric_list = manifest.get(METRICS_STR, [])
        self._plugin_logger = PluginLogger(self.name, self.id)

        self.__class__._logger.info("successfully loaded plugin {}".format(self))

    def __repr__(self):
        return 'Plugin - id: {}, name:{}, is_custom_plugin:{}, profile:{}, manifest:{}'\
            .format(self._id, self._name, self.is_custom_plugin, self._profile, self._manifest)

    @classmethod
    def custom_plugin(cls, script, manifest, id, name):
        manifest = yaml.load(manifest, Loader=yaml.Loader)
        cls._logger.info('custom_plugin id:{}, name:{}, load manifest:{}'.format(id, name, manifest))

        safe_globals = RestrictedModuleLoader(None, None, HubHolder.wf_hub()).get_safe_globals(
            name=name,
            module_list=manifest.get(MODULES, []),
            class_list=manifest.get(CLASSES, [])
        )
        cls._logger.debug('safe_globals:{} for plugin id:{}, name:{}'.format(safe_globals, id, name))

        try:
            # compile script from nestdb
            byte_code = compile_restricted(str(script, 'utf-8'), name, 'exec')
            # TODO only for mock data
            # byte_code = compile_restricted(script, name, 'exec')
        except SyntaxError as se:
            cls._logger.error(
                'Script of plugin has syntax error, pluginId: {}, name:{}, error: {}'.format(id, name, se))
            raise se

        return cls(id, True, safe_globals, byte_code, name, manifest)

    @classmethod
    def predefined_plugin(cls, plugin_path):
        with open(plugin_path.replace('.py', '.manifest.yml'), 'r') as f:
            manifest = yaml.load(f.read(), Loader=yaml.Loader)
            safe_globals = RestrictedModuleLoader(HubHolder.mpa_hub(), HubHolder.alarm_hub(), HubHolder.wf_hub())\
                .get_safe_globals(
                name=os.path.basename(plugin_path),
                module_list=manifest.get(MODULES, []),
                class_list=manifest.get(CLASSES, [])
            )
        cls._logger.debug('safe_globals:{} for plugin_path:{}'.format(safe_globals, plugin_path))

        with open(plugin_path, 'r') as f:
            try:
                byte_code = compile_restricted(f.read(), plugin_path, 'exec')
            except SyntaxError as se:
                cls._logger.error('the script has syntax error: {} {}'.format(plugin_path, se))
                raise se

        return cls(manifest[ID], False, safe_globals, byte_code, PRE_INSTALL_SCRIPTS[plugin_path], manifest)

    def execute(self, metrics):
        Plugin._logger.debug("plugin {} is executing.".format(self._name))
        data_store = {
            'profile': self._profile,
            'host_id': uuid.UUID(ConfigMonitor.host_id()),
            METRICS_STR: metrics
        }

        framework_global = {
            **self._safe_globals,
            'data_store': data_store,
            'logger': self._plugin_logger
        }

        if self.is_custom_plugin:
            framework_global.update({'run_command': self._run_command_proxy})

        exec(self._compiled_script, framework_global)

    @staticmethod
    def _run_command_proxy(cmd, timeout=4):
        """
        Proxy for custom plugin to run cmd
        Default timeout is 4s
        """
        if timeout > 20:
            raise ValueError('Timeout should be no more than 20s.')

        try:
            return run_command(cmd=cmd, shell=False, timeout=timeout, custom=True)
        except Exception as e:
            trace = traceback.format_exc()
            Plugin._logger.error('Failed in _run_command_proxy, cmd:{}, timeout:{}, trace:{}'
                                 .format(cmd, timeout, trace))
            raise e

    def load_profile(self):
        self._profile = Resolver.get(self._id)
        self.__class__._logger.debug("Loaded profile for plugin {}".format(self))
        return self._profile

    def check_pre_condition(self):
        """
        check if pre-condition is reached.
        If all condition are reached return None. Otherwise, return unreached_msg.
        """
        if self.pre_conditions is None:
            return None

        for condition in self.pre_conditions:
            if PRE_CONDITION[condition][REACHED] is False:
                self.__class__._logger.debug("Does not reach pre condition:{}, plugin: {}".format(condition, self))
                return PRE_CONDITION[condition][UNREACHED_MSG]
        return None

    @property
    def name(self):
        return self._name

    @property
    def modules(self):
        return self._safe_globals

    @property
    def id(self):
        return self._id

    @property
    def is_custom_plugin(self):
        return self._is_custom_plugin

    @property
    def profile(self):
        return self._profile

    @property
    def pre_conditions(self):
        return self._pre_conditions

    @property
    def delay_before_first_check(self):
        return self._delay_before_first_check

    @property
    def manifest(self):
        return self._manifest

    @property
    def metric_list(self):
        return self._metric_list
